# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Data handler functions."""

import collections
import datetime
import os
import re
import shlex
import time

import six

try:
  from shlex import quote
except ImportError:
  from pipes import quote

from google.cloud import ndb

from clusterfuzz._internal.base import dates
from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import memoize
from clusterfuzz._internal.base import persistent_cache
from clusterfuzz._internal.base import retry
from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.crash_analysis import crash_analyzer
from clusterfuzz._internal.crash_analysis import severity_analyzer
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.google_cloud_utils import blobs
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell

DATA_BUNDLE_DEFAULT_BUCKET_IAM_ROLE = 'roles/storage.objectAdmin'
DEFAULT_FAIL_RETRIES = 3
DEFAULT_FAIL_WAIT = 1.5
GOMA_DIR_LINE_REGEX = re.compile(r'^\s*goma_dir\s*=')
HEARTBEAT_LAST_UPDATE_KEY = 'heartbeat_update'
INPUT_DIR = 'inputs'
MEMCACHE_TTL_IN_SECONDS = 30 * 60

NUM_TESTCASE_QUALITY_BITS = 3
MAX_TESTCASE_QUALITY = 2**NUM_TESTCASE_QUALITY_BITS - 1

# Value and dimension map for some crash types (timeout, ooms).
CRASH_TYPE_VALUE_REGEX_MAP = {
    'Timeout': r'.*-timeout=(\d+)',
    'Out-of-memory': r'.*-rss_limit_mb=(\d+)',
}
CRASH_TYPE_DIMENSION_MAP = {
    'Timeout': 'secs',
    'Out-of-memory': 'MB',
}

TESTCASE_REPORT_URL = 'https://{domain}/testcase?key={testcase_id}'
TESTCASE_DOWNLOAD_URL = 'https://{domain}/download?testcase_id={testcase_id}'
TESTCASE_REVISION_RANGE_URL = (
    'https://{domain}/revisions?job={job_type}&range={revision_range}')
TESTCASE_REVISION_URL = (
    'https://{domain}/revisions?job={job_type}&revision={revision}')

FILE_UNREPRODUCIBLE_TESTCASE_TEXT = (
    '************************* UNREPRODUCIBLE *************************\n'
    'Note: This crash might not be reproducible with the provided testcase. '
    'That said, for the past %d days, we\'ve been seeing this crash '
    'frequently.\n\n'
    'It may be possible to reproduce by trying the following options:\n'
    '- Run testcase multiple times for a longer duration.\n'
    '- Run fuzzing without testcase argument to hit the same crash signature.\n'
    '\nIf it still does not reproduce, try a speculative fix based on the '
    'crash stacktrace and verify if it works by looking at the crash '
    'statistics in the report. We will auto-close the bug if the crash is not '
    'seen for %d days.\n'
    '******************************************************************' %
    (data_types.FILE_CONSISTENT_UNREPRODUCIBLE_TESTCASE_DEADLINE,
     data_types.UNREPRODUCIBLE_TESTCASE_WITH_BUG_DEADLINE))

FuzzerDisplay = collections.namedtuple(
    'FuzzerDisplay', ['engine', 'target', 'name', 'fully_qualified_name'])

# ------------------------------------------------------------------------------
# Testcase, TestcaseUploadMetadata database related functions
# ------------------------------------------------------------------------------


@memoize.wrap(memoize.Memcache(MEMCACHE_TTL_IN_SECONDS))
def get_all_project_names():
  """Return all project names."""
  query = data_types.Job.query(
      projection=[data_types.Job.project], distinct=True)
  return sorted([job.project for job in query])


def get_domain():
  """Get current domain."""
  default_domain = '{app_id}.appspot.com'.format(
      app_id=utils.get_application_id())
  return local_config.GAEConfig().get('domains.main', default=default_domain)


def get_testcase_by_id(testcase_id):
  """Return the testcase with the given id, or None if it does not exist."""
  if not testcase_id or not str(testcase_id).isdigit() or int(testcase_id) == 0:
    raise errors.InvalidTestcaseError

  testcase = ndb.Key(data_types.Testcase, int(testcase_id)).get()
  if not testcase:
    raise errors.InvalidTestcaseError

  return testcase


def find_testcase(project_name,
                  crash_type,
                  crash_state,
                  security_flag,
                  testcase_to_exclude=None):
  """Find an open test case matching certain parameters."""
  # Prepare the query.
  query = data_types.Testcase.query(
      data_types.Testcase.project_name == project_name,
      data_types.Testcase.crash_type == crash_type,
      data_types.Testcase.crash_state == crash_state,
      data_types.Testcase.security_flag == security_flag,
      data_types.Testcase.status == 'Processed',
      ndb_utils.is_true(data_types.Testcase.open))

  # Return any open (not fixed) test cases if they exist.
  testcases = ndb_utils.get_all_from_query(query)
  testcase = None
  testcase_quality = -1
  for current_testcase in testcases:
    if (testcase_to_exclude and
        current_testcase.key.id() == testcase_to_exclude.key.id()):
      continue
    if current_testcase.duplicate_of:
      continue

    # Replace the current test case in various situations where we have found
    # a better one to use. Testcase quality is based on the following factors:
    # - Is this test case reproducible? Reproducible tests are preferred.
    # - Is there a bug for this? We prefer showing tests with bugs to point
    #   users to existing bugs.
    # - Is this test case minimized ? Minimization confirms that testcase is
    #   reproducible and more usable for reproduction.
    current_testcase_quality = 0
    if not current_testcase.one_time_crasher_flag:
      current_testcase_quality |= 2**2
    if current_testcase.bug_information:
      current_testcase_quality |= 2**1
    if current_testcase.minimized_keys:
      current_testcase_quality |= 2**0

    if current_testcase_quality > testcase_quality:
      testcase = current_testcase
      testcase_quality = current_testcase_quality

    if testcase_quality == MAX_TESTCASE_QUALITY:
      # Already found the best testcase possible, no more work to do. Bail out.
      break

  return testcase


def get_crash_type_string(testcase):
  """Return a crash type string for a testcase."""
  crash_type = ' '.join(testcase.crash_type.splitlines())
  if crash_type not in list(CRASH_TYPE_VALUE_REGEX_MAP.keys()):
    return crash_type

  crash_stacktrace = get_stacktrace(testcase)
  match = re.match(CRASH_TYPE_VALUE_REGEX_MAP[crash_type], crash_stacktrace,
                   re.DOTALL)
  if not match:
    return crash_type

  return '%s (exceeds %s %s)' % (crash_type, match.group(1),
                                 CRASH_TYPE_DIMENSION_MAP[crash_type])


def filter_stacktrace(stacktrace):
  """Filters stacktrace and returns content appropriate for storage as an
  appengine entity."""
  unicode_stacktrace = utils.decode_to_unicode(stacktrace)
  if len(unicode_stacktrace) <= data_types.STACKTRACE_LENGTH_LIMIT:
    return unicode_stacktrace

  tmpdir = environment.get_value('BOT_TMPDIR')
  tmp_stacktrace_file = os.path.join(tmpdir, 'stacktrace.tmp')

  try:
    with open(tmp_stacktrace_file, 'wb') as handle:
      handle.write(unicode_stacktrace.encode('utf-8'))
    with open(tmp_stacktrace_file, 'rb') as handle:
      key = blobs.write_blob(handle)
  except Exception:
    logs.log_error('Unable to write crash stacktrace to temporary file.')
    shell.remove_file(tmp_stacktrace_file)
    return unicode_stacktrace[(-1 * data_types.STACKTRACE_LENGTH_LIMIT):]

  shell.remove_file(tmp_stacktrace_file)
  return '%s%s' % (data_types.BLOBSTORE_STACK_PREFIX, key)


def get_issue_summary(testcase):
  """Gets an issue description string for a testcase."""
  # Get summary prefix. Note that values for fuzzers take priority over those
  # from job definitions.
  fuzzer_summary_prefix = get_value_from_fuzzer_environment_string(
      testcase.fuzzer_name, 'SUMMARY_PREFIX')
  job_summary_prefix = get_value_from_job_definition(testcase.job_type,
                                                     'SUMMARY_PREFIX')
  summary_prefix = fuzzer_summary_prefix or job_summary_prefix or ''

  issue_summary = summary_prefix
  binary_name = testcase.get_metadata('fuzzer_binary_name')
  if binary_name:
    if summary_prefix:
      issue_summary += ':'
    issue_summary += binary_name
  if issue_summary:
    issue_summary += ': '

  # For ASSERTs and CHECK failures, we should just use the crash type and the
  # first line of the crash state as titles. Note that ASSERT_NOT_REACHED should
  # be handled by the general case.
  if testcase.crash_type in [
      'ASSERT', 'CHECK failure', 'Security CHECK failure',
      'Security DCHECK failure'
  ]:
    issue_summary += (
        testcase.crash_type + ': ' + testcase.crash_state.splitlines()[0])
    return issue_summary

  # Special case for bad-cast style testcases.
  if testcase.crash_type == 'Bad-cast':
    filtered_crash_state_lines = testcase.crash_state.splitlines()

    # Add the to/from line (this should always exist).
    issue_summary += filtered_crash_state_lines[0]

    # Add the crash function if available.
    if len(filtered_crash_state_lines) > 1:
      issue_summary += ' in ' + filtered_crash_state_lines[1]

    return issue_summary

  # Add first lines from crash type and crash_state.
  if testcase.crash_type:
    filtered_crash_type = re.sub(r'UNKNOWN( READ| WRITE)?', 'Crash',
                                 testcase.crash_type.splitlines()[0])
    issue_summary += filtered_crash_type
  else:
    issue_summary += 'Unknown error'

  if testcase.crash_state == 'NULL' or not testcase.crash_state:
    # Special case for empty stacktrace.
    issue_summary += ' with empty stacktrace'
  else:
    issue_summary += ' in ' + testcase.crash_state.splitlines()[0]

  return issue_summary


def get_reproduction_help_url(testcase, config):
  """Return url to reproduce the bug."""
  return get_value_from_job_definition_or_environment(
      testcase.job_type, 'HELP_URL', default=config.reproduction_help_url)


def get_fuzzer_display(testcase):
  """Return FuzzerDisplay tuple."""
  if (testcase.overridden_fuzzer_name == testcase.fuzzer_name or
      not testcase.overridden_fuzzer_name):
    return FuzzerDisplay(
        engine=None,
        target=None,
        name=testcase.fuzzer_name,
        fully_qualified_name=testcase.fuzzer_name)

  fuzz_target = get_fuzz_target(testcase.overridden_fuzzer_name)
  if not fuzz_target:
    # Legacy testcases.
    return FuzzerDisplay(
        engine=testcase.fuzzer_name,
        target=testcase.get_metadata('fuzzer_binary_name'),
        name=testcase.fuzzer_name,
        fully_qualified_name=testcase.overridden_fuzzer_name)

  return FuzzerDisplay(
      engine=fuzz_target.engine,
      target=fuzz_target.binary,
      name=fuzz_target.engine,
      fully_qualified_name=fuzz_target.fully_qualified_name())


def filter_arguments(arguments, fuzz_target_name=None):
  """Filter arguments, removing testcase argument and fuzz target binary
  names."""
  # Filter out %TESTCASE*% argument.
  arguments = re.sub(r'[^\s]*%TESTCASE(|_FILE_URL|_HTTP_URL)%', '', arguments)
  if fuzz_target_name:
    arguments = arguments.replace(fuzz_target_name, '')

  return arguments.strip()


def get_arguments(testcase):
  """Return minimized arguments, without testcase argument and fuzz target
  binary itself (for engine fuzzers)."""
  arguments = (
      testcase.minimized_arguments or
      get_value_from_job_definition(testcase.job_type, 'APP_ARGS', default=''))

  # Filter out fuzz target argument. We shouldn't have any case for this other
  # than what is needed by launcher.py for engine based fuzzers.
  fuzzer_display = get_fuzzer_display(testcase)
  fuzz_target = fuzzer_display.target
  return filter_arguments(arguments, fuzz_target)


def _get_memory_tool_options(testcase):
  """Return memory tool options as a string to pass on command line."""
  env = testcase.get_metadata('env')
  if not env:
    return []

  result = []
  for options_name, options_value in sorted(six.iteritems(env)):
    # Strip symbolize flag, use default symbolize=1.
    options_value.pop('symbolize', None)
    if not options_value:
      continue

    options_string = environment.join_memory_tool_options(options_value)
    result.append('{options_name}="{options_string}"'.format(
        options_name=options_name, options_string=quote(options_string)))

  return result


def _get_bazel_test_args(arguments, sanitizer_options):
  """Return arguments to pass to a bazel test."""
  result = []
  for sanitizer_option in sanitizer_options:
    result.append('--test_env=%s' % sanitizer_option)

  for argument in shlex.split(arguments):
    result.append('--test_arg=%s' % quote(argument))

  return ' '.join(result)


def format_issue_information(testcase, format_string):
  """Format a string with information from the testcase."""
  arguments = get_arguments(testcase)
  fuzzer_display = get_fuzzer_display(testcase)
  fuzzer_name = fuzzer_display.name or 'NA'
  fuzz_target = fuzzer_display.target or 'NA'
  engine = fuzzer_display.engine or 'NA'
  last_tested_crash_revision = str(
      testcase.get_metadata('last_tested_crash_revision') or
      testcase.crash_revision)
  project_name = get_project_name(testcase.job_type)
  testcase_id = str(testcase.key.id())
  sanitizer = environment.get_memory_tool_name(testcase.job_type)
  sanitizer_options = _get_memory_tool_options(testcase)
  sanitizer_options_string = ' '.join(sanitizer_options)
  bazel_test_args = _get_bazel_test_args(arguments, sanitizer_options)

  # Multi-target binaries.
  fuzz_target_parts = fuzz_target.split('@')
  base_fuzz_target = fuzz_target_parts[0]
  if len(fuzz_target_parts) == 2:
    fuzz_test_name = fuzz_target_parts[1]
  else:
    fuzz_test_name = ''

  result = format_string.replace('%TESTCASE%', testcase_id)
  result = result.replace('%PROJECT%', project_name)
  result = result.replace('%REVISION%', last_tested_crash_revision)
  result = result.replace('%FUZZER_NAME%', fuzzer_name)
  result = result.replace('%FUZZ_TARGET%', fuzz_target)
  result = result.replace('%BASE_FUZZ_TARGET%', base_fuzz_target)
  result = result.replace('%FUZZ_TEST_NAME%', fuzz_test_name)
  result = result.replace('%ENGINE%', engine)
  result = result.replace('%SANITIZER%', sanitizer)
  result = result.replace('%SANITIZER_OPTIONS%', sanitizer_options_string)
  result = result.replace('%ARGS%', arguments)
  result = result.replace('%BAZEL_TEST_ARGS%', bazel_test_args)
  return result


def get_formatted_reproduction_help(testcase):
  """Return url to reproduce the bug."""
  help_format = get_value_from_job_definition_or_environment(
      testcase.job_type, 'HELP_FORMAT')
  if not help_format:
    return None

  # Since this value may be in a job definition, it's non-trivial for it to
  # include newlines. Instead, it will contain backslash-escaped characters
  # that must be converted here (e.g. \n).
  help_format = help_format.encode().decode('unicode-escape')
  return format_issue_information(testcase, help_format)


def get_plaintext_help_text(testcase, config):
  """Get the help text for this testcase for display in issue descriptions."""
  # Prioritize a HELP_FORMAT message if available.
  formatted_help = get_formatted_reproduction_help(testcase)
  if formatted_help:
    return formatted_help

  # Show a default message and HELP_URL if only it has been supplied.
  help_url = get_reproduction_help_url(testcase, config)
  if help_url:
    return 'See %s for instructions to reproduce this bug locally.' % help_url

  return ''


def get_fixed_range_url(testcase):
  """Return url to testcase fixed range."""
  # Testcase is not fixed yet.
  if not testcase.fixed:
    return None

  # Testcase is unreproducible or coming from a custom binary.
  if testcase.fixed == 'NA' or testcase.fixed == 'Yes':
    return None

  return TESTCASE_REVISION_RANGE_URL.format(
      domain=get_domain(),
      job_type=testcase.job_type,
      revision_range=testcase.fixed)


def get_issue_description(testcase,
                          reporter=None,
                          show_reporter=False,
                          hide_crash_state=False):
  """Returns testcase as string."""
  # Get issue tracker configuration parameters.
  config = db_config.get()
  domain = get_domain()
  testcase_id = testcase.key.id()

  download_url = TESTCASE_DOWNLOAD_URL.format(
      domain=domain, testcase_id=testcase_id)
  report_url = TESTCASE_REPORT_URL.format(
      domain=domain, testcase_id=testcase_id)
  regressed_revision_range_url = TESTCASE_REVISION_RANGE_URL.format(
      domain=domain,
      job_type=testcase.job_type,
      revision_range=testcase.regression)
  revision_range_url = TESTCASE_REVISION_URL.format(
      domain=domain,
      job_type=testcase.job_type,
      revision=testcase.crash_revision)
  fixed_revision_range_url = TESTCASE_REVISION_RANGE_URL.format(
      domain=domain, job_type=testcase.job_type, revision_range=testcase.fixed)

  if testcase.status == 'Unreproducible':
    return ('Testcase {testcase_id} failed to reproduce the crash. '
            'Please inspect the program output at {report_url}.'.format(
                testcase_id=testcase_id, report_url=report_url))

  # Now create the content string.
  content_string = 'Detailed Report: %s\n\n' % report_url

  project_name = get_project_name(testcase.job_type)
  if project_name and project_name != utils.default_project_name():
    content_string += 'Project: %s\n' % project_name

  fuzzer_display = get_fuzzer_display(testcase)
  if fuzzer_display.engine:
    content_string += 'Fuzzing Engine: %s\n' % fuzzer_display.engine
    content_string += 'Fuzz Target: %s\n' % fuzzer_display.target
  else:
    content_string += 'Fuzzer: %s\n' % fuzzer_display.name

  content_string += 'Job Type: %s\n' % testcase.job_type

  # Add platform id if other than default ones. Only applicable to Android.
  # e.g. android:shamu_asan
  if testcase.platform_id:
    content_string += 'Platform Id: %s\n\n' % testcase.platform_id

  content_string += 'Crash Type: %s\n' % get_crash_type_string(testcase)
  content_string += 'Crash Address: %s\n' % testcase.crash_address

  if hide_crash_state:
    crash_state = '...see report...'
  else:
    crash_state = testcase.crash_state
  content_string += 'Crash State:\n%s\n' % (
      utils.indent_string(crash_state + '\n', 2))

  content_string += '%s\n\n' % environment.get_memory_tool_display_string(
      testcase.job_type)

  if data_types.SecuritySeverity.is_valid(testcase.security_severity):
    content_string += (
        'Recommended Security Severity: %s\n\n' %
        severity_analyzer.severity_to_string(testcase.security_severity))

  if (testcase.regression and testcase.regression != 'NA' and
      not testcase.regression.startswith('0:') and
      not testcase.regression.endswith('!')):
    content_string += 'Regressed: %s\n' % regressed_revision_range_url
  else:
    content_string += 'Crash Revision: %s\n' % revision_range_url

  if (testcase.fixed and testcase.fixed != 'NA' and testcase.fixed != 'Yes' and
      not testcase.fixed.endswith('!')):
    content_string += 'Fixed: %s\n' % fixed_revision_range_url

  if not content_string.endswith('\n\n'):
    content_string += '\n'

  content_string += 'Reproducer Testcase: %s\n\n' % download_url

  if testcase.gestures:
    content_string += 'Additional requirements: Requires Gestures\n\n'
  if testcase.http_flag:
    content_string += 'Additional requirements: Requires HTTP\n\n'

  if show_reporter:
    if reporter:
      content_string += (
          'Issue manually filed by: %s\n\n' % reporter.split('@')[0])
    else:
      content_string += 'Issue filed automatically.\n\n'

  # Jobs can override the help url.
  content_string += get_plaintext_help_text(testcase, config)

  # Unreproducible crash text is only applicable when we are consistently seeing
  # it happening, and hence the reason for auto-filing it. Otherwise, someone
  # filed it manually, so skip the text in that case.
  if not reporter and testcase.one_time_crasher_flag:
    content_string += '\n\n' + FILE_UNREPRODUCIBLE_TESTCASE_TEXT

  # Add additional body text from metadata.
  issue_metadata = testcase.get_metadata('issue_metadata', {})
  additional_fields = issue_metadata.get('additional_fields', {})
  additional_fields_strs = []
  for key, value in additional_fields.items():
    additional_fields_strs.append(f'{key}: {value}')
  if additional_fields_strs:
    content_string += '\n\n' + '\n'.join(additional_fields_strs)

  return content_string


def get_stacktrace(testcase, stack_attribute='crash_stacktrace'):
  """Returns the stacktrace for a test case.

  This may require a blobstore read.
  """
  result = getattr(testcase, stack_attribute)
  if not result or not result.startswith(data_types.BLOBSTORE_STACK_PREFIX):
    return result

  # For App Engine, we can't write to local file, so use blobs.read_key instead.
  if environment.is_running_on_app_engine():
    key = result[len(data_types.BLOBSTORE_STACK_PREFIX):]
    return str(blobs.read_key(key), 'utf-8', errors='replace')

  key = result[len(data_types.BLOBSTORE_STACK_PREFIX):]
  tmpdir = environment.get_value('BOT_TMPDIR')
  tmp_stacktrace_file = os.path.join(tmpdir, 'stacktrace.tmp')
  blobs.read_blob_to_disk(key, tmp_stacktrace_file)

  try:
    with open(tmp_stacktrace_file) as handle:
      result = handle.read()
  except:
    logs.log_error(
        'Unable to read stacktrace for testcase %d.' % testcase.key.id())
    result = ''

  shell.remove_file(tmp_stacktrace_file)
  return result


def handle_duplicate_entry(testcase):
  """Handles duplicates and deletes unreproducible one."""
  # Caller ensures that our testcase object is up-to-date. If someone else
  # already marked us as a duplicate, no more work to do.
  if testcase.duplicate_of:
    return

  existing_testcase = find_testcase(
      testcase.project_name,
      testcase.crash_type,
      testcase.crash_state,
      testcase.security_flag,
      testcase_to_exclude=testcase)
  if not existing_testcase:
    return

  # If the existing testcase's minimization has not completed yet, we shouldn't
  # be doing the next step. The testcase might turn out to be a non reproducible
  # bug and we don't want to delete the other testcase which could be a fully
  # minimized and reproducible bug.
  if not existing_testcase.minimized_keys:
    return

  testcase_id = testcase.key.id()
  existing_testcase_id = existing_testcase.key.id()
  if (not testcase.bug_information and
      not existing_testcase.one_time_crasher_flag):
    metadata = data_types.TestcaseUploadMetadata.query(
        data_types.TestcaseUploadMetadata.testcase_id == testcase_id).get()
    if metadata:
      metadata.status = 'Duplicate'
      metadata.duplicate_of = existing_testcase_id
      metadata.security_flag = existing_testcase.security_flag
      metadata.put()

    testcase.status = 'Duplicate'
    testcase.duplicate_of = existing_testcase_id
    testcase.put()
    logs.log('Marking testcase %d as duplicate of testcase %d.' %
             (testcase_id, existing_testcase_id))

  elif (not existing_testcase.bug_information and
        not testcase.one_time_crasher_flag):
    metadata = data_types.TestcaseUploadMetadata.query(
        data_types.TestcaseUploadMetadata.testcase_id == testcase_id).get()
    if metadata:
      metadata.status = 'Duplicate'
      metadata.duplicate_of = testcase_id
      metadata.security_flag = testcase.security_flag
      metadata.put()

    existing_testcase.status = 'Duplicate'
    existing_testcase.duplicate_of = testcase_id
    existing_testcase.put()
    logs.log('Marking testcase %d as duplicate of testcase %d.' %
             (existing_testcase_id, testcase_id))


def is_first_retry_for_task(testcase, reset_after_retry=False):
  """Returns true if this task is tried atleast once. Only applicable for
  analyze and progression tasks."""
  task_name = environment.get_value('TASK_NAME')
  retry_key = '%s_retry' % task_name
  retry_flag = testcase.get_metadata(retry_key)
  if not retry_flag:
    # Update the metadata key since now we have tried it once.
    retry_value = True
    testcase.set_metadata(retry_key, retry_value)
    return True

  # Reset the metadata key so that tasks like progression task can be retried.
  if reset_after_retry:
    retry_value = False
    testcase.set_metadata(retry_key, retry_value)

  return False


@memoize.wrap(memoize.Memcache(MEMCACHE_TTL_IN_SECONDS))
def get_issue_tracker_name(job_type=None):
  """Return issue tracker name for a job type."""
  return get_value_from_job_definition_or_environment(job_type, 'ISSUE_TRACKER')


@memoize.wrap(memoize.Memcache(MEMCACHE_TTL_IN_SECONDS))
def get_project_name(job_type):
  """Return project name for a job type."""
  default_project_name = utils.default_project_name()
  return get_value_from_job_definition(job_type, 'PROJECT_NAME',
                                       default_project_name)


@memoize.wrap(memoize.Memcache(MEMCACHE_TTL_IN_SECONDS))
def get_main_repo(job_type):
  """Return project name for a job type."""
  return get_value_from_job_definition(job_type, 'MAIN_REPO')


def _get_security_severity(crash, job_type, gestures):
  """Get security severity."""
  if crash.security_flag:
    return severity_analyzer.get_security_severity(
        crash.crash_type, crash.crash_stacktrace, job_type, bool(gestures))

  return None


def store_testcase(crash, fuzzed_keys, minimized_keys, regression, fixed,
                   one_time_crasher_flag, crash_revision, comment,
                   absolute_path, fuzzer_name, fully_qualified_fuzzer_name,
                   job_type, archived, archive_filename, binary_flag, http_flag,
                   gestures, redzone, disable_ubsan, minidump_keys,
                   window_argument, timeout_multiplier, minimized_arguments):
  """Create a testcase and store it in the datastore using remote api."""
  # Initialize variable to prevent invalid values.
  if archived:
    archive_state = data_types.ArchiveStatus.FUZZED
  else:
    archive_state = 0
  if not gestures:
    gestures = []
  if not redzone:
    redzone = 128

  # Create the testcase.
  testcase = data_types.Testcase()
  testcase.crash_type = crash.crash_type
  testcase.crash_address = crash.crash_address
  testcase.crash_state = utils.decode_to_unicode(crash.crash_state)
  testcase.crash_stacktrace = filter_stacktrace(crash.crash_stacktrace)
  testcase.fuzzed_keys = fuzzed_keys
  testcase.minimized_keys = minimized_keys
  testcase.bug_information = ''
  testcase.regression = regression
  testcase.fixed = fixed
  testcase.security_flag = crash.security_flag
  testcase.security_severity = _get_security_severity(crash, job_type, gestures)

  testcase.one_time_crasher_flag = one_time_crasher_flag
  testcase.crash_revision = crash_revision
  testcase.absolute_path = absolute_path
  testcase.fuzzer_name = fuzzer_name
  testcase.overridden_fuzzer_name = fully_qualified_fuzzer_name or fuzzer_name
  testcase.job_type = job_type
  testcase.queue = tasks.default_queue()
  testcase.archive_state = archive_state
  testcase.archive_filename = archive_filename
  testcase.binary_flag = binary_flag
  testcase.http_flag = http_flag
  testcase.timestamp = datetime.datetime.utcnow()
  testcase.gestures = gestures
  testcase.redzone = redzone
  testcase.disable_ubsan = disable_ubsan
  testcase.minidump_keys = minidump_keys
  testcase.window_argument = window_argument
  testcase.timeout_multiplier = float(timeout_multiplier)
  testcase.minimized_arguments = minimized_arguments
  testcase.project_name = get_project_name(job_type)

  # Set metadata fields (e.g. build url, build key, platform string, etc).
  set_initial_testcase_metadata(testcase)

  # Update the comment and save testcase.
  update_testcase_comment(testcase, data_types.TaskState.NA, comment)

  # Get testcase id from newly created testcase.
  testcase_id = testcase.key.id()
  logs.log(
      ('Created new testcase %d (reproducible:%s, security:%s, binary:%s).\n'
       'crash_type: %s\ncrash_state:\n%s\n') %
      (testcase_id, not testcase.one_time_crasher_flag, testcase.security_flag,
       testcase.binary_flag, testcase.crash_type, testcase.crash_state))

  # Update global blacklist to avoid finding this leak again (if needed).
  is_lsan_enabled = environment.get_value('LSAN')
  if is_lsan_enabled:
    from clusterfuzz._internal.fuzzing import leak_blacklist
    leak_blacklist.add_crash_to_global_blacklist_if_needed(testcase)

  return testcase_id


def set_initial_testcase_metadata(testcase):
  """Set various testcase metadata fields during testcase initialization."""
  build_key = environment.get_value('BUILD_KEY')
  if build_key:
    testcase.set_metadata('build_key', build_key, update_testcase=False)

  build_url = environment.get_value('BUILD_URL')
  if build_url:
    testcase.set_metadata('build_url', build_url, update_testcase=False)

  gn_args_path = environment.get_value('GN_ARGS_PATH', '')
  if gn_args_path and os.path.exists(gn_args_path):
    gn_args = utils.read_data_from_file(
        gn_args_path, eval_data=False, default='').decode('utf-8')

    # Remove goma_dir from gn args since it is only relevant to the machine that
    # did the build.
    filtered_gn_args_lines = [
        line for line in gn_args.splitlines()
        if not GOMA_DIR_LINE_REGEX.match(line)
    ]
    filtered_gn_args = '\n'.join(filtered_gn_args_lines)
    testcase.set_metadata('gn_args', filtered_gn_args, update_testcase=False)

  testcase.platform = environment.platform().lower()
  testcase.platform_id = environment.get_platform_id()


def update_testcase_comment(testcase, task_state, message=None):
  """Add task status and message to the test case's comment field."""
  bot_name = environment.get_value('BOT_NAME', 'Unknown')
  task_name = environment.get_value('TASK_NAME', 'Unknown')
  task_string = '%s task' % task_name.capitalize()
  timestamp = utils.current_date_time()

  # For some tasks like blame, progression and impact, we need to delete lines
  # from old task executions to avoid clutter.
  if (task_name in ['blame', 'progression', 'impact'] and
      task_state == data_types.TaskState.STARTED):
    pattern = r'.*?: %s.*\n' % task_string
    testcase.comments = re.sub(pattern, '', testcase.comments)

  testcase.comments += '[%s] %s: %s %s' % (timestamp, bot_name, task_string,
                                           task_state)
  if message:
    testcase.comments += ': %s' % message.rstrip('.')
  testcase.comments += '.\n'

  # Truncate if too long.
  if len(testcase.comments) > data_types.TESTCASE_COMMENTS_LENGTH_LIMIT:
    logs.log_error(
        'Testcase comments truncated (testcase {testcase_id}, job {job_type}).'.
        format(testcase_id=testcase.key.id(), job_type=testcase.job_type))
    testcase.comments = testcase.comments[
        -data_types.TESTCASE_COMMENTS_LENGTH_LIMIT:]

  testcase.put()

  # Log the message in stackdriver after the testcase.put() call as otherwise
  # the testcase key might not available yet (i.e. for new testcase).
  if message:
    log_func = (
        logs.log_error
        if task_state == data_types.TaskState.ERROR else logs.log)
    log_func('{message} (testcase {testcase_id}, job {job_type}).'.format(
        message=message,
        testcase_id=testcase.key.id(),
        job_type=testcase.job_type))


def get_open_testcase_id_iterator():
  """Get an iterator for open testcase ids."""
  keys = ndb_utils.get_all_from_query(
      data_types.Testcase.query(
          ndb_utils.is_true(data_types.Testcase.open),
          data_types.Testcase.status == 'Processed'),
      keys_only=True,
      batch_size=data_types.TESTCASE_ENTITY_QUERY_LIMIT)
  for key in keys:
    yield key.id()


def critical_tasks_completed(testcase):
  """Check to see if all critical tasks have finished running on a test case."""
  if testcase.status == 'Unreproducible':
    # These tasks don't apply to unreproducible testcases.
    return True

  if testcase.one_time_crasher_flag:
    # These tasks don't apply to flaky testcases.
    return True

  # For non-chromium projects, impact and blame tasks are not applicable.
  if not utils.is_chromium():
    return testcase.minimized_keys and testcase.regression

  return bool(testcase.minimized_keys and testcase.regression and
              testcase.is_impact_set_flag)


# ------------------------------------------------------------------------------
# BuildMetadata database related functions
# ------------------------------------------------------------------------------


def get_build_state(job_type, crash_revision):
  """Return whether a build is unmarked, good or bad."""
  build = data_types.BuildMetadata.query(
      data_types.BuildMetadata.job_type == job_type,
      data_types.BuildMetadata.revision == crash_revision).get()

  if not build:
    return data_types.BuildState.UNMARKED

  if build.bad_build:
    return data_types.BuildState.BAD

  return data_types.BuildState.GOOD


def add_build_metadata(job_type,
                       crash_revision,
                       is_bad_build,
                       console_output=None):
  """Add build metadata."""
  build = data_types.BuildMetadata()
  build.bad_build = is_bad_build
  build.bot_name = environment.get_value('BOT_NAME')
  build.console_output = filter_stacktrace(console_output)
  build.job_type = job_type
  build.revision = crash_revision
  build.timestamp = datetime.datetime.utcnow()
  build.put()

  if is_bad_build:
    logs.log_error(
        'Bad build %s.' % job_type,
        revision=crash_revision,
        job_type=job_type,
        output=console_output)
  else:
    logs.log(
        'Good build %s.' % job_type, revision=crash_revision, job_type=job_type)
  return build


# Fuzzer and DataBundle database related functions
# ------------------------------------------------------------------------------


def create_data_bundle_bucket_and_iams(data_bundle_name, emails):
  """Creates a data bundle bucket and adds iams for access."""
  bucket_name = get_data_bundle_bucket_name(data_bundle_name)
  if not storage.create_bucket_if_needed(bucket_name):
    return False

  client = storage.create_discovery_storage_client()
  iam_policy = storage.get_bucket_iam_policy(client, bucket_name)
  if not iam_policy:
    return False

  members = []

  # Add access for the domains allowed in project.
  domains = local_config.AuthConfig().get('whitelisted_domains', default=[])
  for domain in domains:
    members.append('domain:%s' % domain)

  # Add access for the emails provided in function arguments.
  for email in emails:
    members.append('user:%s' % email)

  if not members:
    # No members to add, bail out.
    return True

  binding = storage.get_bucket_iam_binding(iam_policy,
                                           DATA_BUNDLE_DEFAULT_BUCKET_IAM_ROLE)
  if binding:
    binding['members'] = members
  else:
    binding = {
        'role': DATA_BUNDLE_DEFAULT_BUCKET_IAM_ROLE,
        'members': members,
    }
    iam_policy['bindings'].append(binding)

  return bool(storage.set_bucket_iam_policy(client, bucket_name, iam_policy))


def bucket_domain_suffix():
  domain = local_config.ProjectConfig().get('bucket_domain_suffix')
  if not domain:
    domain = '%s.appspot.com' % utils.get_application_id()

  return domain


def get_data_bundle_bucket_name(data_bundle_name):
  """Return data bundle bucket name on GCS."""
  domain = bucket_domain_suffix()
  return '%s-corpus.%s' % (data_bundle_name, domain)


def get_data_bundle_bucket_url(data_bundle_name):
  """Return data bundle bucket url on GCS."""
  return 'gs://%s' % get_data_bundle_bucket_name(data_bundle_name)


def get_value_from_fuzzer_environment_string(fuzzer_name,
                                             variable_pattern,
                                             default=None):
  """Get a specific environment variable's value for a fuzzer."""
  fuzzer = data_types.Fuzzer.query(data_types.Fuzzer.name == fuzzer_name).get()
  if not fuzzer or not fuzzer.additional_environment_string:
    return default

  return get_value_from_environment_string(
      fuzzer.additional_environment_string, variable_pattern, default=default)


# ------------------------------------------------------------------------------
# TaskStatus database related functions
# ------------------------------------------------------------------------------


def get_task_status(name, create_if_needed=False):
  """Return the TaskStatus object with the given name."""
  metadata = ndb.Key(data_types.TaskStatus, name).get()
  if not metadata and create_if_needed:
    metadata = data_types.TaskStatus(id=name)

  return metadata


def update_task_status(task_name, status, expiry_interval=None):
  """Updates status for a task. Used to ensure that a single instance of a task
  is running at any given time."""
  bot_name = environment.get_value('BOT_NAME')
  failure_wait_interval = environment.get_value('FAIL_WAIT')

  # If we didn't get an expiry interval, default to our task lease interval.
  if expiry_interval is None:
    expiry_interval = environment.get_value('TASK_LEASE_SECONDS')
    if expiry_interval is None:
      logs.log_error('expiry_interval is None and TASK_LEASE_SECONDS not set.')

  def _try_update_status():
    """Try update metadata."""
    task_status = get_task_status(task_name, create_if_needed=True)

    # If another bot is already working on this task, bail out with error.
    if (status == data_types.TaskState.STARTED and
        task_status.status == data_types.TaskState.STARTED and
        not dates.time_has_expired(
            task_status.time, seconds=expiry_interval - 1)):
      return False

    task_status.bot_name = bot_name
    task_status.status = status
    task_status.time = utils.utcnow()
    task_status.put()
    return True

  # It is important that we do not continue until the metadata is updated.
  # This can lead to task loss, or can cause issues with multiple bots
  # attempting to run the task at the same time.
  while True:
    try:
      return ndb.transaction(_try_update_status, retries=0)
    except Exception:
      # We need to update the status under all circumstances.
      # Failing to update 'completed' status causes another bot
      # that picked up this job to bail out.
      logs.log_error('Unable to update %s task metadata. Retrying.' % task_name)
      time.sleep(utils.random_number(1, failure_wait_interval))


# ------------------------------------------------------------------------------
# Heartbeat database related functions
# ------------------------------------------------------------------------------


def update_heartbeat(force_update=False):
  """Updates heartbeat with current timestamp and log data."""
  # Check if the heartbeat was recently updated. If yes, bail out.
  last_modified_time = persistent_cache.get_value(
      HEARTBEAT_LAST_UPDATE_KEY, constructor=datetime.datetime.utcfromtimestamp)
  if (not force_update and last_modified_time and not dates.time_has_expired(
      last_modified_time, seconds=data_types.HEARTBEAT_WAIT_INTERVAL)):
    return 0

  bot_name = environment.get_value('BOT_NAME')
  current_time = datetime.datetime.utcnow()

  try:
    heartbeat = ndb.Key(data_types.Heartbeat, bot_name).get()
    if not heartbeat:
      heartbeat = data_types.Heartbeat()
      heartbeat.bot_name = bot_name

    heartbeat.key = ndb.Key(data_types.Heartbeat, bot_name)
    heartbeat.task_payload = tasks.get_task_payload()
    heartbeat.task_end_time = tasks.get_task_end_time()
    heartbeat.last_beat_time = current_time
    heartbeat.source_version = utils.current_source_version()
    heartbeat.platform_id = environment.get_platform_id()
    heartbeat.put()

    persistent_cache.set_value(
        HEARTBEAT_LAST_UPDATE_KEY, time.time(), persist_across_reboots=True)
  except:
    logs.log_error('Unable to update heartbeat.')
    return 0

  return 1


def bot_run_timed_out():
  """Return true if our run timed out."""
  run_timeout = environment.get_value('RUN_TIMEOUT')
  if not run_timeout:
    return False

  start_time = environment.get_value('START_TIME')
  if not start_time:
    return False

  start_time = datetime.datetime.utcfromtimestamp(start_time)

  # Actual run timeout takes off the duration for one task.
  average_task_duration = environment.get_value('AVERAGE_TASK_DURATION', 0)
  actual_run_timeout = run_timeout - average_task_duration

  return dates.time_has_expired(start_time, seconds=actual_run_timeout)


# ------------------------------------------------------------------------------
# Job database related functions
# ------------------------------------------------------------------------------


@memoize.wrap(memoize.Memcache(MEMCACHE_TTL_IN_SECONDS))
def get_component_name(job_type):
  """Gets component name for a job type."""
  job = data_types.Job.query(data_types.Job.name == job_type).get()
  if not job:
    return ''

  match = re.match(r'.*BUCKET_PATH[^\r\n]*-([a-zA-Z0-9]+)-component',
                   job.get_environment_string(), re.DOTALL)
  if not match:
    return ''

  component_name = match.group(1)
  return component_name


@memoize.wrap(memoize.Memcache(MEMCACHE_TTL_IN_SECONDS))
def get_repository_for_component(component):
  """Get the repository based on component."""
  default_repository = ''
  repository = ''
  repository_mappings = db_config.get_value('component_repository_mappings')

  for line in repository_mappings.splitlines():
    current_component, value = line.split(';', 1)

    if current_component == 'default':
      default_repository = value
    elif current_component == component:
      repository = value

  return repository or default_repository


def get_value_from_environment_string(environment_string,
                                      variable_pattern,
                                      default=None):
  """Return the first value matching the pattern from the environment string."""
  pattern = r'%s\s*=\s*(.*)' % variable_pattern
  match = re.search(pattern, environment_string)
  if not match:
    return default

  return match.group(1).strip()


def get_value_from_job_definition(job_type, variable_pattern, default=None):
  """Get a specific environment variable's value from a job definition."""
  if not job_type:
    return default

  job = data_types.Job.query(data_types.Job.name == job_type).get()
  if not job:
    return default

  return job.get_environment().get(variable_pattern, default)


def get_value_from_job_definition_or_environment(job_type,
                                                 variable_pattern,
                                                 default=None):
  """Gets a specific environment variable's value from a job definition. If
  not found, it returns the value from current environment."""
  return get_value_from_job_definition(
      job_type,
      variable_pattern,
      default=environment.get_value(variable_pattern, default))


def get_additional_values_for_variable(variable_name, job_type, fuzzer_name):
  """Helper function to read a list of additional items from a job definition
     and fuzzer's additional environment string."""
  value_list_strings = [
      get_value_from_job_definition(job_type, variable_name),
      get_value_from_fuzzer_environment_string(fuzzer_name, variable_name),
  ]

  additional_values = []
  for value_list_string in value_list_strings:
    if value_list_string:
      # Ignore whitespace between commas.
      additional_values += [v.strip() for v in value_list_string.split(',')]

  return additional_values


# ------------------------------------------------------------------------------
# Notification database related functions
# ------------------------------------------------------------------------------


def is_notification_sent(testcase_id, user_email):
  """Return true if this notification has already been sent."""
  notification = data_types.Notification.query(
      data_types.Notification.testcase_id == testcase_id,
      data_types.Notification.user_email == user_email).get()
  return bool(notification)


def create_notification_entry(testcase_id, user_email):
  """Create a entry log for sent notification."""
  notification = data_types.Notification()
  notification.testcase_id = testcase_id
  notification.user_email = user_email
  notification.put()


# ------------------------------------------------------------------------------
# TestcaseUploadMetadata database related functions
# ------------------------------------------------------------------------------


def create_user_uploaded_testcase(key,
                                  original_key,
                                  archive_state,
                                  filename,
                                  file_path_input,
                                  timeout,
                                  job,
                                  queue,
                                  http_flag,
                                  gestures,
                                  additional_arguments,
                                  bug_information,
                                  crash_revision,
                                  uploader_email,
                                  platform_id,
                                  app_launch_command,
                                  fuzzer_name,
                                  fully_qualified_fuzzer_name,
                                  fuzzer_binary_name,
                                  bundled,
                                  retries,
                                  bug_summary_update_flag,
                                  quiet_flag,
                                  additional_metadata=None,
                                  crash_data=None):
  """Create a testcase object, metadata, and task for a user uploaded test."""
  testcase = data_types.Testcase()
  if crash_data:
    # External job with provided stacktrace.
    testcase.crash_type = crash_data.crash_type
    testcase.crash_state = crash_data.crash_state
    testcase.crash_address = crash_data.crash_address
    testcase.crash_stacktrace = crash_data.crash_stacktrace

    testcase.status = 'Processed'
    testcase.security_flag = crash_analyzer.is_security_issue(
        testcase.crash_stacktrace, testcase.crash_type, testcase.crash_address)
    testcase.regression = 'NA'
    testcase.comments = '[%s] %s: External testcase upload.\n' % (
        utils.current_date_time(), uploader_email)
    # External jobs never get minimized.
    testcase.minimized_keys = 'NA'

    # analyze_task sets this for non-external reproductions.
    testcase.platform = job.platform.lower()
    testcase.platform_id = testcase.platform
  else:
    testcase.crash_type = ''
    testcase.crash_state = 'Pending'
    testcase.crash_address = ''
    testcase.crash_stacktrace = ''
    testcase.status = 'Pending'
    testcase.security_flag = False
    testcase.regression = ''
    testcase.comments = '[%s] %s: Analyze task.\n' % (utils.current_date_time(),
                                                      uploader_email)
    testcase.minimized_keys = ''

  testcase.fuzzed_keys = key
  testcase.bug_information = ''
  testcase.fixed = ''
  testcase.one_time_crasher_flag = False
  testcase.crash_revision = crash_revision
  testcase.fuzzer_name = fuzzer_name
  testcase.overridden_fuzzer_name = fully_qualified_fuzzer_name or fuzzer_name
  testcase.job_type = job.name
  testcase.http_flag = bool(http_flag)
  testcase.archive_state = archive_state
  testcase.project_name = get_project_name(job.name)

  if archive_state or bundled:
    testcase.absolute_path = file_path_input
    testcase.archive_filename = filename
  else:
    testcase.absolute_path = filename
  testcase.gestures = gestures
  if bug_information and bug_information.isdigit() and int(bug_information):
    testcase.bug_information = bug_information
  if platform_id:
    testcase.platform_id = platform_id.strip().lower()
  if additional_arguments:
    testcase.set_metadata(
        'uploaded_additional_args', additional_arguments, update_testcase=False)
  if app_launch_command:
    testcase.set_metadata(
        'app_launch_command', app_launch_command, update_testcase=False)
  if fuzzer_binary_name:
    testcase.set_metadata(
        'fuzzer_binary_name', fuzzer_binary_name, update_testcase=False)

  if additional_metadata:
    for metadata_key, metadata_value in six.iteritems(additional_metadata):
      testcase.set_metadata(metadata_key, metadata_value, update_testcase=False)

  testcase.timestamp = utils.utcnow()
  testcase.uploader_email = uploader_email
  testcase.put()

  # Store the testcase upload metadata.
  testcase_id = testcase.key.id()
  metadata = data_types.TestcaseUploadMetadata()
  metadata.security_flag = testcase.security_flag
  metadata.filename = filename
  if testcase.status == 'Processed':
    metadata.status = 'Confirmed'
  else:
    metadata.status = 'Pending'

  metadata.uploader_email = uploader_email
  metadata.testcase_id = testcase_id
  metadata.blobstore_key = key
  metadata.original_blobstore_key = original_key
  metadata.timeout = timeout
  metadata.bundled = bundled
  metadata.retries = retries
  if bundled:
    metadata.path_in_archive = file_path_input
  metadata.timestamp = testcase.timestamp
  metadata.bug_summary_update_flag = bug_summary_update_flag
  metadata.quiet_flag = quiet_flag
  metadata.bug_information = testcase.bug_information

  if crash_data:
    if crash_analyzer.ignore_stacktrace(testcase.crash_stacktrace):
      close_invalid_uploaded_testcase(testcase, metadata, 'Irrelevant')
      return testcase.key.id()

    if check_uploaded_testcase_duplicate(testcase, metadata):
      close_invalid_uploaded_testcase(testcase, metadata, 'Duplicate')
      return testcase.key.id()

  metadata.put()

  # Create the job to analyze the testcase.
  tasks.add_task('analyze', testcase_id, job.name, queue)
  return testcase.key.id()


def check_uploaded_testcase_duplicate(testcase, metadata):
  """Check if the uploaded testcase is a duplicate."""
  existing_testcase = find_testcase(testcase.project_name, testcase.crash_type,
                                    testcase.crash_state,
                                    testcase.security_flag)

  if not existing_testcase or existing_testcase.key.id() == testcase.key.id():
    return False

  # If the existing test case is unreproducible and we are, replace the
  # existing test case with this one.
  if (existing_testcase.one_time_crasher_flag and
      not testcase.one_time_crasher_flag):
    duplicate_testcase = existing_testcase
    original_testcase = testcase
  else:
    duplicate_testcase = testcase
    original_testcase = existing_testcase
    metadata.status = 'Duplicate'
    metadata.duplicate_of = existing_testcase.key.id()

  duplicate_testcase.status = 'Duplicate'
  duplicate_testcase.duplicate_of = original_testcase.key.id()
  duplicate_testcase.put()

  return duplicate_testcase.key.id() == testcase.key.id()


def close_invalid_uploaded_testcase(testcase, metadata, status):
  """Closes an invalid testcase and updates metadata."""
  testcase.status = status
  testcase.open = False
  testcase.minimized_keys = 'NA'
  testcase.regression = 'NA'
  testcase.set_impacts_as_na()
  testcase.fixed = 'NA'
  testcase.triaged = True
  testcase.put()

  metadata.status = status
  metadata.put()


# ------------------------------------------------------------------------------
# TestcaseGroup related functions
# ------------------------------------------------------------------------------


def delete_group(group_id, update_testcases=True):
  """Delete the testcase group with the specified id if it exists."""
  # Remove all testcases from the group.
  if update_testcases:
    testcases = get_testcases_in_group(group_id)
    for testcase in testcases:
      remove_testcase_from_group(testcase)

  # Delete the group itself.
  group = get_entity_by_type_and_id(data_types.TestcaseGroup, group_id)
  if group:
    group.key.delete()


def get_testcase_ids_in_group(group_id):
  """Return the all testcase ids in the specified group."""
  if not group_id or not str(group_id).isdigit():
    return []

  query = ndb_utils.get_all_from_query(
      data_types.Testcase.query(data_types.Testcase.group_id == int(group_id)),
      keys_only=True)
  return [key.id() for key in query]


def get_testcases_in_group(group_id):
  """Return the all testcases in the specified group."""
  # Fetch by keys (strongly consistent) to avoid stale results from query
  # (eventually consistent).
  testcases = []
  for testcase_id in get_testcase_ids_in_group(group_id):
    try:
      testcases.append(get_testcase_by_id(testcase_id))
    except errors.InvalidTestcaseError:
      # Already deleted.
      continue

  return testcases


def remove_testcase_from_group(testcase):
  """Removes a testcase from group."""
  if not testcase:
    return

  testcase.group_id = 0
  testcase.group_bug_information = 0
  testcase.put()


def update_group_bug(group_id):
  """Update group bug information for a group."""
  if not group_id:
    # No associated group, no work to do. Bail out.
    return

  testcases = get_testcases_in_group(group_id)
  if not testcases:
    # No group members found. Bail out.
    return

  group_bug_information = 0
  for testcase in testcases:
    if not testcase.bug_information:
      continue

    issue_id = int(testcase.bug_information)
    if not group_bug_information:
      group_bug_information = issue_id
    else:
      group_bug_information = min(group_bug_information, issue_id)

  for testcase in testcases:
    testcase.group_bug_information = group_bug_information
  ndb_utils.put_multi(testcases)


# ------------------------------------------------------------------------------
# Generic helper functions for any data type
# ------------------------------------------------------------------------------


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='datastore.data_handler.get_entity_by_type_and_id')
def get_entity_by_type_and_id(entity_type, entity_id):
  """Return the datastore object with the given type and id if it exists."""
  if not entity_id or not str(entity_id).isdigit() or int(entity_id) == 0:
    return None

  return entity_type.get_by_id(int(entity_id))


# ------------------------------------------------------------------------------
# TestcaseVariant related functions
# ------------------------------------------------------------------------------


def get_testcase_variant(testcase_id, job_type):
  """Get a testcase variant entity, and create if needed."""
  testcase_id = int(testcase_id)
  variant = data_types.TestcaseVariant.query(
      data_types.TestcaseVariant.testcase_id == testcase_id,
      data_types.TestcaseVariant.job_type == job_type).get()
  if not variant:
    variant = data_types.TestcaseVariant(
        testcase_id=testcase_id, job_type=job_type)
  return variant


# ------------------------------------------------------------------------------
# Fuzz target related functions
# ------------------------------------------------------------------------------

FUZZ_TARGET_UPDATE_FAIL_RETRIES = 5
FUZZ_TARGET_UPDATE_FAIL_DELAY = 2


@retry.wrap(
    retries=FUZZ_TARGET_UPDATE_FAIL_RETRIES,
    delay=FUZZ_TARGET_UPDATE_FAIL_DELAY,
    function='datastore.data_handler.record_fuzz_target')
def record_fuzz_target(engine_name, binary_name, job_type):
  """Record existence of fuzz target."""
  if not binary_name:
    logs.log_error('Expected binary_name.')
    return None

  project = get_project_name(job_type)
  key_name = data_types.fuzz_target_fully_qualified_name(
      engine_name, project, binary_name)

  fuzz_target = ndb.Key(data_types.FuzzTarget, key_name).get()
  if not fuzz_target:
    fuzz_target = data_types.FuzzTarget(
        engine=engine_name, project=project, binary=binary_name)
    fuzz_target.put()

  job_mapping_key = data_types.fuzz_target_job_key(key_name, job_type)
  job_mapping = ndb.Key(data_types.FuzzTargetJob, job_mapping_key).get()
  if job_mapping:
    job_mapping.last_run = utils.utcnow()
  else:
    job_mapping = data_types.FuzzTargetJob(
        fuzz_target_name=key_name,
        job=job_type,
        engine=engine_name,
        last_run=utils.utcnow())
  job_mapping.put()

  logs.log(
      'Recorded use of fuzz target %s.' % key_name,
      project=project,
      engine=engine_name,
      binary_name=binary_name,
      job_type=job_type)
  return fuzz_target


def get_fuzz_target(name):
  """Get FuzzTarget by fully qualified name."""
  if not name:
    return None

  return ndb.Key(data_types.FuzzTarget, name).get()


def get_fuzz_target_job(fuzz_target_name, job):
  """Get FuzzTargetJob by fully qualified name and job."""
  return ndb.Key(data_types.FuzzTargetJob,
                 data_types.fuzz_target_job_key(fuzz_target_name, job)).get()


def get_fuzz_targets(engine=None, project=None, binary=None):
  """Return a Datastore query for fuzz targets."""
  query = data_types.FuzzTarget().query()

  if engine:
    query = query.filter(data_types.FuzzTarget.engine == engine)

  if project:
    query = query.filter(data_types.FuzzTarget.project == project)

  if binary:
    query = query.filter(data_types.FuzzTarget.binary == binary)

  return ndb_utils.get_all_from_query(query)


def get_fuzzing_engines():
  """Return the fuzzing engines currently running."""
  query = data_types.FuzzTarget.query(
      projection=[data_types.FuzzTarget.engine], distinct=True)
  return [f.engine for f in ndb_utils.get_all_from_query(query)]


def is_fuzzing_engine(name):
  """Return whether or not |name| is a fuzzing engine."""
  query = data_types.FuzzTarget.query(data_types.FuzzTarget.engine == name)
  return bool(query.count(limit=1))


@memoize.wrap(memoize.Memcache(MEMCACHE_TTL_IN_SECONDS))
def get_all_fuzzer_names_including_children(include_parents=False,
                                            project=None):
  """Returns all fuzzer names, including expanded child fuzzers."""
  all_fuzzers = set()
  engine_fuzzers = get_fuzzing_engines()

  fuzzers = data_types.Fuzzer.query(projection=['name'])
  for fuzzer in fuzzers:
    # Add this if we're including all parents or this is not an engine fuzzer
    # with fuzz targets.
    if include_parents or fuzzer.name not in engine_fuzzers:
      all_fuzzers.add(fuzzer.name)

  for fuzz_target in get_fuzz_targets(project=project):
    all_fuzzers.add(fuzz_target.fully_qualified_name())

  return sorted(list(all_fuzzers))


@memoize.wrap(memoize.Memcache(MEMCACHE_TTL_IN_SECONDS))
def get_all_job_type_names(project=None):
  """Return all job type names."""
  query = data_types.Job.query(projection=['name'])
  if project:
    query = query.filter(data_types.Job.project == project)
  return sorted([job.name for job in query])


def get_coverage_information(fuzzer_name, date, create_if_needed=False):
  """Get coverage information, or create if it doesn't exist."""
  coverage_info = ndb.Key(
      data_types.CoverageInformation,
      data_types.coverage_information_key(fuzzer_name, date)).get()

  if not coverage_info and create_if_needed:
    coverage_info = data_types.CoverageInformation(
        fuzzer=fuzzer_name, date=date)

  return coverage_info


def close_testcase_with_error(testcase_id, error_message):
  """Close testcase (fixed=NA) with an error message."""
  testcase = get_testcase_by_id(testcase_id)
  update_testcase_comment(testcase, data_types.TaskState.ERROR, error_message)
  testcase.fixed = 'NA'
  testcase.open = False
  testcase.put()


def clear_progression_pending(testcase):
  """If we marked progression as pending for this testcase, clear that state."""
  if not testcase.get_metadata('progression_pending'):
    return

  testcase.delete_metadata('progression_pending', update_testcase=False)


def update_progression_completion_metadata(testcase,
                                           revision,
                                           is_crash=False,
                                           message=None):
  """Update metadata the progression task completes."""
  clear_progression_pending(testcase)
  testcase.set_metadata('last_tested_revision', revision, update_testcase=False)
  if is_crash:
    testcase.set_metadata(
        'last_tested_crash_revision', revision, update_testcase=False)
    testcase.set_metadata(
        'last_tested_crash_time', utils.utcnow(), update_testcase=False)
  if not testcase.open:
    testcase.set_metadata('closed_time', utils.utcnow(), update_testcase=False)
  update_testcase_comment(testcase, data_types.TaskState.FINISHED, message)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for testcase management."""

import base64
import collections
import datetime
import os
import re
import zlib

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.build_management import revisions
from clusterfuzz._internal.crash_analysis import crash_analyzer
from clusterfuzz._internal.crash_analysis.crash_comparer import CrashComparer
from clusterfuzz._internal.crash_analysis.crash_result import CrashResult
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_init
from clusterfuzz._internal.metrics import fuzzer_logs
from clusterfuzz._internal.metrics import fuzzer_stats
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.platforms import android
from clusterfuzz._internal.system import archive
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import process_handler
from clusterfuzz._internal.system import shell
from clusterfuzz.fuzz import engine

# Testcase filename prefixes and suffixes.
CRASH_PREFIX = 'crash-'
FUZZ_PREFIX = 'fuzz-'
FLAGS_PREFIX = 'flags-'
HTTP_PREFIX = 'http-'
RESOURCES_PREFIX = 'resources-'

# TODO(mbarbella): Once all fuzzers are converted to "resources-", remove this.
DEPENDENCY_PREFIX = 'cfdependency-'
APPS_PREFIX = 'fuzz-apps-'
EXTENSIONS_PREFIX = 'fuzz-extension-'
COVERAGE_SUFFIX = '.cov'

INFO_FILE_EXTENSION = '.info'
IPCDUMP_EXTENSION = '.ipcdump'
REPRODUCIBILITY_FACTOR = 0.5
SEARCH_INDEX_TESTCASES_DIRNAME = 'common'
SEARCH_INDEX_BUNDLE_PREFIX = '__%s_' % SEARCH_INDEX_TESTCASES_DIRNAME
TESTCASE_LIST_FILENAME = 'files.info'

CHROME_URL_LOAD_REGEX = re.compile(
    r'.*(NetworkDelegate::NotifyBeforeURLRequest|FileURLLoader::Start)'
    r':\s+(.*)')
FILE_URL_REGEX = re.compile(r'file:///([^"#?]+)')
HTTP_URL_REGEX = re.compile(
    r'.*(localhost|\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})[^/]*[/]([^"#?]+)')

BAD_STATE_HINTS = [
    # X server issues.
    'cannot open display',
    'Maximum number of clients reached',
    'Missing X server',

    # Android logging issues.
    'logging service has stopped',
]


class TestcaseManagerError(Exception):
  """Base exception."""


class TargetNotFoundError(TestcaseManagerError):
  """Error when a fuzz target is not found."""


def create_testcase_list_file(output_directory):
  """Create a testcase list file for tests in a directory."""
  files_list = []
  files_list_file_path = os.path.join(output_directory, TESTCASE_LIST_FILENAME)
  for root, _, files in shell.walk(output_directory):
    for filename in files:
      if filename.endswith(INFO_FILE_EXTENSION):
        # Skip an info file.
        continue

      file_path = os.path.join(root, filename)
      if not utils.is_valid_testcase_file(file_path, check_if_exists=False):
        continue

      normalized_relative_file_path = utils.get_normalized_relative_path(
          file_path, output_directory)
      files_list.append(normalized_relative_file_path)

  utils.write_data_to_file('\n'.join(sorted(files_list)), files_list_file_path)


def get_testcases_from_directories(directories):
  """Returns all testcases from testcase directories."""
  testcase_paths = []
  max_testcases = environment.get_value('MAX_TESTCASES')

  generators = []
  for directory in directories:
    if not directory.strip():
      continue

    generators.append(shell.walk(directory))

  for generator in generators:
    for structure in generator:
      base_directory = structure[0]
      for filename in structure[2]:
        if not filename.startswith(FUZZ_PREFIX):
          continue

        if filename.endswith(COVERAGE_SUFFIX):
          continue

        file_path = os.path.join(base_directory, filename)
        if not os.path.getsize(file_path):
          continue

        testcase_paths.append(utils.normalize_path(file_path))
        if len(testcase_paths) == max_testcases:
          return testcase_paths

  return testcase_paths


def is_testcase_resource(filename):
  """Returns true if this is a testcase or its resource dependency."""
  if filename.startswith(FUZZ_PREFIX):
    return True

  if filename.startswith(FLAGS_PREFIX):
    return True

  if filename.startswith(DEPENDENCY_PREFIX):
    return True

  if filename.startswith(RESOURCES_PREFIX):
    return True

  if filename.endswith(COVERAGE_SUFFIX):
    return True

  return False


def remove_testcases_from_directories(directories):
  """Removes all testcases and their dependencies from testcase directories."""
  generators = []
  for directory in directories:
    if not directory.strip():
      continue

    # If there is a bot-specific files list, delete it now.
    bot_testcases_file_path = utils.get_bot_testcases_file_path(directory)
    shell.remove_file(bot_testcases_file_path)

    generators.append(shell.walk(directory))

  for generator in generators:
    for structure in generator:
      base_directory = structure[0]
      for filename in structure[2]:
        if not is_testcase_resource(filename):
          continue

        if filename.startswith(RESOURCES_PREFIX):
          # In addition to removing this file, remove all resources.
          resources_file_path = os.path.join(base_directory, filename)
          resources = read_resource_list(resources_file_path)
          for resource in resources:
            shell.remove_file(resource)

        file_path = os.path.join(base_directory, filename)
        shell.remove_file(file_path)


def read_resource_list(resource_file_path):
  """Generate a resource list."""
  if not os.path.exists(resource_file_path):
    return []

  resources = []
  base_directory = os.path.dirname(resource_file_path)
  with open(resource_file_path) as file_handle:
    resource_file_contents = file_handle.read()
    for line in resource_file_contents.splitlines():
      resource = os.path.join(base_directory, line.strip())
      if not os.path.exists(resource):
        break

      resources.append(resource)

  return resources


def get_resource_dependencies(testcase_absolute_path, test_prefix=FUZZ_PREFIX):
  """Returns the list of testcase resource dependencies."""
  resources = []
  if not os.path.exists(testcase_absolute_path):
    return resources

  base_directory = os.path.dirname(testcase_absolute_path)
  testcase_filename = os.path.basename(testcase_absolute_path)

  # FIXME(mbarbella): Remove this when all fuzzers are using "resources-".
  # This code includes the dependencies that begin with
  # dependency prefix and are referenced in the testcase.
  testcase_contents = None
  for filename in os.listdir(base_directory):
    if filename.startswith(DEPENDENCY_PREFIX):
      # Only load the testcase contents if necessary.
      if not testcase_contents:
        with open(testcase_absolute_path, 'rb') as file_handle:
          testcase_contents = file_handle.read()

      if filename.encode('utf-8') in testcase_contents:
        file_path = os.path.join(base_directory, filename)
        resources.append(file_path)

  # This code includes the dependencies in cases when the testcase itself is a
  # just a wrapper file around the actual testcase.
  if DEPENDENCY_PREFIX in testcase_absolute_path:
    dependency_filename = os.path.splitext(testcase_filename)[0]
    dependency_filename = re.compile(DEPENDENCY_PREFIX).sub(
        '', dependency_filename, 1)
    dependency_filename = re.compile(FUZZ_PREFIX).sub('', dependency_filename,
                                                      1)
    dependency_filename = re.compile(HTTP_PREFIX).sub('', dependency_filename,
                                                      1)
    dependency_file_path = os.path.join(base_directory, dependency_filename)
    resources.append(dependency_file_path)

  # Check to see if this test case lists all resources in a resources file.
  if testcase_filename.startswith(test_prefix):
    stripped_testcase_name = testcase_filename[len(test_prefix):]
    resources_filename = '%s%s' % (RESOURCES_PREFIX, stripped_testcase_name)
    resources_file_path = os.path.join(base_directory, resources_filename)
    resources += read_resource_list(resources_file_path)

  # For extensions, archive everything in the extension directory.
  if APPS_PREFIX in testcase_filename or EXTENSIONS_PREFIX in testcase_filename:
    for root, _, files in shell.walk(base_directory):
      for filename in files:
        file_path = os.path.join(root, filename)
        if file_path == testcase_absolute_path:
          continue

        resources.append(file_path)

  return resources


def get_command_line_flags(testcase_path):
  """Returns command line flags to use for a testcase."""
  arguments = environment.get_value('APP_ARGS')
  additional_arguments = get_additional_command_line_flags(testcase_path)
  if arguments:
    arguments += ' ' + additional_arguments
  else:
    arguments = additional_arguments

  return arguments.strip()


def get_additional_command_line_flags(testcase_path):
  """Returns additional command line flags to use for a testcase."""
  # Get the initial flags list from the environment value.
  additional_command_line_flags = (
      environment.get_value('ADDITIONAL_COMMAND_LINE_FLAGS', ''))

  # If we don't have a fuzz prefix, no need to look further for flags file.
  testcase_filename = os.path.basename(testcase_path)
  if not testcase_filename.startswith(FUZZ_PREFIX):
    return additional_command_line_flags

  # Gets the flags list from the flags file.
  stripped_testcase_name = testcase_filename[len(FUZZ_PREFIX):]
  flags_filename = '%s%s' % (FLAGS_PREFIX, stripped_testcase_name)
  flags_file_path = os.path.join(os.path.dirname(testcase_path), flags_filename)
  flags_file_content = utils.read_data_from_file(
      flags_file_path, eval_data=False)
  if flags_file_content:
    additional_command_line_flags += ' ' + flags_file_content.decode('utf-8')
  return additional_command_line_flags.strip()


def run_testcase(thread_index, file_path, gestures, env_copy):
  """Run a single testcase and return crash results in the crash queue."""
  try:
    # Update environment with environment copy from parent.
    if env_copy:
      os.environ.update(env_copy)

    # Initialize variables.
    needs_http = '-http-' in file_path
    test_timeout = environment.get_value('TEST_TIMEOUT', 10)
    app_directory = environment.get_value('APP_DIR')
    environment.set_value('PIDS', '[]')

    # Get command line options.
    command = get_command_line_for_application(
        file_path, user_profile_index=thread_index, needs_http=needs_http)

    # Run testcase.
    return process_handler.run_process(
        command,
        timeout=test_timeout,
        gestures=gestures,
        env_copy=env_copy,
        current_working_directory=app_directory)
  except Exception:
    logs.log_error('Exception occurred while running run_testcase.')

    return None, None, None


class Crash(
    collections.namedtuple(
        'Crash', 'file_path crash_time return_code resource_list gestures '
        'stack_file_path')):
  """Represents a crash in a queue. This class is transformed into
    fuzz_task.Crash. Therefore, please be careful when adding/removing
    fields."""


def get_resource_paths(output):
  """Read the urls from the output."""
  resource_paths = set()
  for line in output.splitlines():
    match = CHROME_URL_LOAD_REGEX.match(line)
    if not match:
      continue

    local_path = convert_dependency_url_to_local_path(match.group(2))
    if local_path:
      logs.log('Detected resource: %s.' % local_path)
      resource_paths.add(local_path)

  return list(resource_paths)


def convert_dependency_url_to_local_path(url):
  """Convert a dependency URL to a corresponding local path."""
  # Bot-specific import.
  from clusterfuzz._internal.bot.webserver import http_server

  logs.log('Process dependency: %s.' % url)
  file_match = FILE_URL_REGEX.search(url)
  http_match = HTTP_URL_REGEX.search(url)
  platform = environment.platform()

  local_path = None
  if file_match:
    file_path = file_match.group(1)
    logs.log('Detected file dependency: %s.' % file_path)
    if platform == 'WINDOWS':
      local_path = file_path
    else:
      local_path = '/' + file_path

      # Convert remote to local path for android.
      if environment.is_android():
        remote_testcases_directory = android.constants.DEVICE_TESTCASES_DIR
        local_testcases_directory = environment.get_value('FUZZ_INPUTS')
        local_path = local_path.replace(remote_testcases_directory,
                                        local_testcases_directory)

  elif http_match:
    relative_http_path = os.path.sep + http_match.group(2)
    logs.log('Detected http dependency: %s.' % relative_http_path)
    local_path = http_server.get_absolute_testcase_file(relative_http_path)
    if not local_path:
      # This needs to be a warning since in many cases, it is actually a
      # non-existent path. For others, we need to add the directory aliases in
      # file http_server.py.
      logs.log_warn(
          'Unable to find server resource %s, skipping.' % relative_http_path)

  if local_path:
    local_path = utils.normalize_path(local_path)

  return local_path


def _get_testcase_time(testcase_path):
  """Returns the timestamp of a testcase."""
  stats = fuzzer_stats.TestcaseRun.read_from_disk(testcase_path)
  if stats:
    return datetime.datetime.utcfromtimestamp(float(stats.timestamp))

  return None


def upload_testcase(testcase_path, log_time):
  """Uploads testcase so that a log file can be matched with it folder."""
  fuzz_logs_bucket = environment.get_value('FUZZ_LOGS_BUCKET')
  if not fuzz_logs_bucket:
    return

  if not os.path.exists(testcase_path):
    return

  with open(testcase_path, 'rb') as file_handle:
    testcase_contents = file_handle.read()

  fuzzer_logs.upload_to_logs(
      fuzz_logs_bucket,
      testcase_contents,
      time=log_time,
      file_extension='.testcase')


def _get_crash_output(output):
  """Returns crash part of the output, excluding unrelated content (e.g. output
  from corpus merge, etc)."""
  if output is None:
    return None

  crash_stacktrace_end_marker_index = output.find(
      data_types.CRASH_STACKTRACE_END_MARKER)
  if crash_stacktrace_end_marker_index == -1:
    return output

  return output[:crash_stacktrace_end_marker_index]


def run_testcase_and_return_result_in_queue(crash_queue,
                                            thread_index,
                                            file_path,
                                            gestures,
                                            env_copy,
                                            upload_output=False):
  """Run a single testcase and return crash results in the crash queue."""
  # Since this is running in its own process, initialize the log handler again.
  # This is needed for Windows where instances are not shared across child
  # processes. See:
  # https://stackoverflow.com/questions/34724643/python-logging-with-multiprocessing-root-logger-different-in-windows
  logs.configure('run_testcase', {
      'testcase_path': file_path,
  })

  # Also reinitialize NDB context for the same reason as above.
  with ndb_init.context():
    _do_run_testcase_and_return_result_in_queue(
        crash_queue,
        thread_index,
        file_path,
        gestures,
        env_copy,
        upload_output=upload_output)


def _do_run_testcase_and_return_result_in_queue(crash_queue,
                                                thread_index,
                                                file_path,
                                                gestures,
                                                env_copy,
                                                upload_output=False):
  """Run a single testcase and return crash results in the crash queue."""
  try:
    # Run testcase and check whether a crash occurred or not.
    return_code, crash_time, output = run_testcase(thread_index, file_path,
                                                   gestures, env_copy)

    # Pull testcase directory to host to get any stats files.
    if environment.is_trusted_host():
      from clusterfuzz._internal.bot.untrusted_runner import file_host
      file_host.pull_testcases_from_worker()

    # Analyze the crash.
    crash_output = _get_crash_output(output)
    crash_result = CrashResult(return_code, crash_time, crash_output)

    # To provide consistency between stats and logs, we use timestamp taken
    # from stats when uploading logs and testcase.
    if upload_output:
      log_time = _get_testcase_time(file_path)

    if crash_result.is_crash():
      # Initialize resource list with the testcase path.
      resource_list = [file_path]
      resource_list += get_resource_paths(crash_output)

      # Store the crash stack file in the crash stacktrace directory
      # with filename as the hash of the testcase path.
      crash_stacks_directory = environment.get_value('CRASH_STACKTRACES_DIR')
      stack_file_path = os.path.join(crash_stacks_directory,
                                     utils.string_hash(file_path))
      utils.write_data_to_file(crash_output, stack_file_path)

      # Put crash/no-crash results in the crash queue.
      crash_queue.put(
          Crash(
              file_path=file_path,
              crash_time=crash_time,
              return_code=return_code,
              resource_list=resource_list,
              gestures=gestures,
              stack_file_path=stack_file_path))

      # Don't upload uninteresting testcases (no crash) or if there is no log to
      # correlate it with (not upload_output).
      if upload_output:
        upload_testcase(file_path, log_time)

    if upload_output:
      # Include full output for uploaded logs (crash output, merge output, etc).
      crash_result_full = CrashResult(return_code, crash_time, output)
      log = prepare_log_for_upload(crash_result_full.get_stacktrace(),
                                   return_code)
      upload_log(log, log_time)
  except Exception:
    logs.log_error('Exception occurred while running '
                   'run_testcase_and_return_result_in_queue.')


def engine_reproduce(engine_impl, target_name, testcase_path, arguments,
                     timeout):
  """Do engine reproduction."""
  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import tasks_host
    return tasks_host.engine_reproduce(engine_impl, target_name, testcase_path,
                                       arguments, timeout)
  build_dir = environment.get_value('BUILD_DIR')
  is_blackbox = engine_impl.name == 'blackbox'
  target_path = engine_common.find_fuzzer_path(
      build_dir, target_name, is_blackbox=is_blackbox)
  if not target_path:
    raise TargetNotFoundError('Failed to find target ' + target_name)

  result = engine_impl.reproduce(target_path, testcase_path, list(arguments),
                                 timeout)

  # This matches the check in process_handler.run_process.
  if not result.return_code and \
      (crash_analyzer.is_memory_tool_crash(result.output) or
       crash_analyzer.is_check_failure_crash(result.output)):
    result.return_code = 1

  return result


class TestcaseRunner(object):
  """Testcase runner."""

  def __init__(self,
               fuzz_target,
               testcase_path,
               test_timeout,
               gestures,
               needs_http=False,
               arguments=None):
    self._testcase_path = testcase_path
    self._test_timeout = test_timeout
    self._gestures = gestures
    self._needs_http = needs_http

    if fuzz_target:
      engine_impl = engine.get(fuzz_target.engine)
    else:
      engine_impl = None

    # TODO(ochang): Make this hard fail once migration to new fuzzing pipeline
    # is complete.
    if fuzz_target and engine_impl:
      self._is_black_box = False
      self._engine_impl = engine_impl

      # Read target_name + args.
      if not arguments:
        arguments = get_command_line_flags(testcase_path)

      arguments = data_handler.filter_arguments(arguments, fuzz_target.binary)
      self._arguments = arguments.split()

      self._fuzz_target = fuzz_target
    else:
      self._is_black_box = True
      self._command = get_command_line_for_application(
          testcase_path, needs_http=needs_http)

  def run(self, round_number):
    """Run the testcase once."""
    app_directory = environment.get_value('APP_DIR')
    warmup_timeout = environment.get_value('WARMUP_TIMEOUT')
    run_timeout = warmup_timeout if round_number == 1 else self._test_timeout

    if self._is_black_box:
      return_code, crash_time, output = process_handler.run_process(
          self._command,
          timeout=run_timeout,
          gestures=self._gestures,
          current_working_directory=app_directory)
    else:
      try:
        result = engine_reproduce(self._engine_impl, self._fuzz_target.binary,
                                  self._testcase_path, self._arguments,
                                  run_timeout)
      except TimeoutError:
        # Treat reproduction timeouts as not crashing.
        return CrashResult(0, run_timeout, '')

      return_code = result.return_code
      crash_time = result.time_executed

      log_header = engine_common.get_log_header(result.command,
                                                result.time_executed)
      output = log_header + '\n' + result.output

    process_handler.terminate_stale_application_instances()

    crash_result = CrashResult(return_code, crash_time, output)
    if not crash_result.is_crash():
      logs.log(
          'No crash occurred (round {round_number}).'.format(
              round_number=round_number),
          output=output)

    return crash_result

  def _pre_run_cleanup(self):
    """Common cleanup before running a testcase."""
    # Cleanup any existing application instances and user profile directories.
    # Cleaning up temp user profile directories. Should be done before calling
    # |get_command_line_for_application| call since that creates dependencies in
    # the profile folder.
    process_handler.terminate_stale_application_instances()
    shell.clear_temp_directory()

  def _get_crash_state(self, round_number, crash_result):
    """Get crash state from a CrashResult."""
    state = crash_result.get_symbolized_data()
    if crash_result.is_crash():
      logs.log(
          ('Crash occurred in {crash_time} seconds (round {round_number}). '
           'State:\n{crash_state}').format(
               crash_time=crash_result.crash_time,
               round_number=round_number,
               crash_state=state.crash_state),
          output=state.crash_stacktrace)

    return state

  def reproduce_with_retries(self,
                             retries,
                             expected_state=None,
                             expected_security_flag=None,
                             flaky_stacktrace=False):
    """Try reproducing a crash with retries."""
    self._pre_run_cleanup()
    crash_result = None

    for round_number in range(1, retries + 1):
      crash_result = self.run(round_number)
      state = self._get_crash_state(round_number, crash_result)

      if not crash_result.is_crash():
        continue

      if not expected_state:
        logs.log('Crash stacktrace comparison skipped.')
        return crash_result

      if crash_result.should_ignore():
        logs.log('Crash stacktrace matched ignore signatures, ignored.')
        continue

      if crash_result.is_security_issue() != expected_security_flag:
        logs.log('Crash security flag does not match, ignored.')
        continue

      if flaky_stacktrace:
        logs.log('Crash stacktrace is marked flaky, skipping comparison.')
        return crash_result

      crash_comparer = CrashComparer(state.crash_state, expected_state)
      if crash_comparer.is_similar():
        logs.log('Crash stacktrace is similar to original stacktrace.')
        return crash_result
      logs.log('Crash stacktrace does not match original stacktrace.')

    logs.log('Didn\'t crash at all.')
    return CrashResult(return_code=0, crash_time=0, output=crash_result.output)

  def test_reproduce_reliability(self, retries, expected_state,
                                 expected_security_flag):
    """Test to see if a crash is fully reproducible or is a one-time crasher."""
    self._pre_run_cleanup()

    reproducible_crash_target_count = retries * REPRODUCIBILITY_FACTOR
    round_number = 0
    crash_count = 0
    for round_number in range(1, retries + 1):
      # Bail out early if there is no hope of finding a reproducible crash.
      if (retries - round_number + crash_count + 1 <
          reproducible_crash_target_count):
        break

      crash_result = self.run(round_number)
      state = self._get_crash_state(round_number, crash_result)

      if not crash_result.is_crash():
        continue

      # If we don't have an expected crash state, set it to the one from initial
      # crash.
      if not expected_state:
        expected_state = state.crash_state

      if crash_result.is_security_issue() != expected_security_flag:
        logs.log('Detected a crash without the correct security flag.')
        continue

      crash_comparer = CrashComparer(state.crash_state, expected_state)
      if not crash_comparer.is_similar():
        logs.log(
            'Detected a crash with an unrelated state: '
            'Expected(%s), Found(%s).' % (expected_state, state.crash_state))
        continue

      crash_count += 1
      if crash_count >= reproducible_crash_target_count:
        logs.log('Crash is reproducible.')
        return True

    logs.log('Crash is not reproducible. Crash count: %d/%d.' % (crash_count,
                                                                 round_number))
    return False


def test_for_crash_with_retries(testcase,
                                testcase_path,
                                test_timeout,
                                http_flag=False,
                                use_gestures=True,
                                compare_crash=True,
                                crash_retries=None):
  """Test for a crash and return crash parameters like crash type, crash state,
  crash stacktrace, etc."""
  gestures = testcase.gestures if use_gestures else None
  try:
    fuzz_target = testcase.get_fuzz_target()
    if engine.get(testcase.fuzzer_name) and not fuzz_target:
      raise TargetNotFoundError

    runner = TestcaseRunner(fuzz_target, testcase_path, test_timeout, gestures,
                            http_flag)

    if crash_retries is None:
      crash_retries = environment.get_value('CRASH_RETRIES')

    if compare_crash:
      expected_state = testcase.crash_state
      expected_security_flag = testcase.security_flag
    else:
      expected_state = None
      expected_security_flag = None

    return runner.reproduce_with_retries(crash_retries, expected_state,
                                         expected_security_flag,
                                         testcase.flaky_stack)
  except TargetNotFoundError:
    # If a target isn't found, treat it as not crashing.
    return CrashResult(return_code=0, crash_time=0, output='')


def test_for_reproducibility(fuzzer_name,
                             full_fuzzer_name,
                             testcase_path,
                             expected_state,
                             expected_security_flag,
                             test_timeout,
                             http_flag,
                             gestures,
                             arguments=None):
  """Test to see if a crash is fully reproducible or is a one-time crasher."""
  try:
    fuzz_target = data_handler.get_fuzz_target(full_fuzzer_name)
    if engine.get(fuzzer_name) and not fuzz_target:
      raise TargetNotFoundError

    runner = TestcaseRunner(
        fuzz_target,
        testcase_path,
        test_timeout,
        gestures,
        http_flag,
        arguments=arguments)

    crash_retries = environment.get_value('CRASH_RETRIES')
    return runner.test_reproduce_reliability(crash_retries, expected_state,
                                             expected_security_flag)
  except TargetNotFoundError:
    # If a target isn't found, treat it as not crashing.
    return False


def prepare_log_for_upload(symbolized_output, return_code):
  """Prepare log for upload."""
  # Add revision information to the logs.
  app_revision = environment.get_value('APP_REVISION')
  job_name = environment.get_value('JOB_NAME')
  components = revisions.get_component_list(app_revision, job_name)
  component_revisions = (
      revisions.format_revision_list(components, use_html=False) or
      'Not available.\n')

  revisions_header =\
  f'Component revisions (build r{app_revision}):\n{component_revisions}\n'

  bot_name = environment.get_value('BOT_NAME')
  bot_header = f'Bot name: {bot_name}\n'
  if environment.is_android():
    bot_header += f'Device serial: {environment.get_value("ANDROID_SERIAL")}\n'

  return_code_header = "Return code: %s\n\n" % return_code

  result = revisions_header + bot_header + return_code_header +\
  symbolized_output
  return result.encode('utf-8')


def upload_log(log, log_time):
  """Upload the output into corresponding GCS logs bucket."""
  fuzz_logs_bucket = environment.get_value('FUZZ_LOGS_BUCKET')
  if not fuzz_logs_bucket:
    return

  fuzzer_logs.upload_to_logs(fuzz_logs_bucket, log, time=log_time)


def get_user_profile_directory(user_profile_index):
  """Returns a user profile directory from a directory index."""
  temp_directory = environment.get_value('BOT_TMPDIR')
  user_profile_in_memory = environment.get_value('USER_PROFILE_IN_MEMORY')
  user_profile_root_directory = (
      temp_directory if user_profile_in_memory else
      environment.get_value('USER_PROFILE_ROOT_DIR'))

  # Create path to user profile directory.
  user_profile_directory_name = 'user_profile_%d' % user_profile_index
  user_profile_directory = os.path.join(user_profile_root_directory,
                                        user_profile_directory_name)

  return user_profile_directory


def get_command_line_for_application(file_to_run='',
                                     user_profile_index=0,
                                     app_path=None,
                                     app_args=None,
                                     needs_http=False,
                                     write_command_line_file=False,
                                     get_arguments_only=False):
  """Returns the complete command line required to execute application."""
  if app_args is None:
    app_args = environment.get_value('APP_ARGS')
  if app_path is None:
    app_path = environment.get_value('APP_PATH')

  if not app_path:
    # No APP_PATH is available for e.g. grey box fuzzers.
    return ''

  additional_command_line_flags = get_additional_command_line_flags(file_to_run)
  app_args_append_testcase = environment.get_value('APP_ARGS_APPEND_TESTCASE')
  app_directory = environment.get_value('APP_DIR')
  app_name = environment.get_value('APP_NAME')
  apps_argument = environment.get_value('APPS_ARG')
  crash_stacks_directory = environment.get_value('CRASH_STACKTRACES_DIR')
  debugger = environment.get_value('DEBUGGER_PATH')
  device_testcases_directory = android.constants.DEVICE_TESTCASES_DIR
  fuzzer_directory = environment.get_value('FUZZER_DIR')
  extension_argument = environment.get_value('EXTENSION_ARG')
  input_directory = environment.get_value('INPUT_DIR')
  launcher = environment.get_value('LAUNCHER_PATH')
  is_android = environment.is_android()
  root_directory = environment.get_value('ROOT_DIR')
  temp_directory = environment.get_value('BOT_TMPDIR')
  user_profile_argument = environment.get_value('USER_PROFILE_ARG')
  window_argument = environment.get_value('WINDOW_ARG')
  user_profile_directory = get_user_profile_directory(user_profile_index)

  # Create user profile directory and setup contents if needed.
  setup_user_profile_directory_if_needed(user_profile_directory)

  # Handle spaces in APP_PATH.
  # If application path has spaces, then we need to quote it.
  if ' ' in app_path:
    app_path = '"%s"' % app_path

  interpreter = shell.get_interpreter(app_name)
  if get_arguments_only:
    # If we are only returning the arguments, do not return the application
    # path or anything else required to run it such as an interpreter.
    app_path = ''
  elif interpreter:
    # Prepend command with interpreter if it is a script.
    app_path = '%s %s' % (interpreter, app_path)

  # Start creating the command line.
  command = ''

  # Rebase the file_to_run and launcher paths to the worker's root.
  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import file_host
    file_to_run = file_host.rebase_to_worker_root(file_to_run)
    launcher = file_host.rebase_to_worker_root(launcher)

  # Default case.
  testcase_path = file_to_run
  testcase_filename = os.path.basename(testcase_path)
  testcase_directory = os.path.dirname(testcase_path)
  testcase_file_url = utils.file_path_to_file_url(testcase_path)
  testcase_http_url = ''

  # Determine where |testcase_file_url| should point depending on platform and
  # whether or not a launcher script is used.
  if file_to_run:
    if launcher:
      # In the case of launcher scripts, the testcase file to be run resides on
      # the host running the launcher script. Thus |testcase_file_url|, which
      # may point to a location on the device for Android job types, does not
      # apply. Instead, the launcher script should be passed the original file
      # to run. By setting |testcase_file_url| to |file_to_run|, we avoid
      # duplicating job definitions solely for supporting launcher scripts.
      testcase_file_url = file_to_run
      # Jobs that have a launcher script which needs to be run on the host will
      # have app_name == launcher. In this case don't prepend launcher to
      # command - just use app_name.
      if os.path.basename(launcher) != app_name:
        launcher_with_interpreter = shell.get_execute_command(launcher)
        command += launcher_with_interpreter + ' '
    elif is_android:
      # Android-specific testcase path fixup for fuzzers that don't rely on
      # launcher scripts.
      local_testcases_directory = environment.get_value('FUZZ_INPUTS')

      # Check if the file to run is in fuzzed testcases folder. If yes, then we
      # can substitute with a local device path. Otherwise, it is part of some
      # data bundle with resource dependencies and we just need to use http
      # host forwarder for that.
      if file_to_run.startswith(local_testcases_directory):
        testcase_relative_path = (
            file_to_run[len(local_testcases_directory) + 1:])
        testcase_path = os.path.join(device_testcases_directory,
                                     testcase_relative_path)
        testcase_file_url = utils.file_path_to_file_url(testcase_path)
      else:
        # Force use of host_forwarder based on comment above.
        needs_http = True

    # Check if the testcase needs to be loaded over http.
    # TODO(ochang): Make this work for trusted/untrusted.
    http_ip = '127.0.0.1'
    http_port_1 = environment.get_value('HTTP_PORT_1', 8000)
    relative_testcase_path = file_to_run[len(input_directory + os.path.sep):]
    relative_testcase_path = relative_testcase_path.replace('\\', '/')
    testcase_http_url = 'http://%s:%d/%s' % (http_ip, http_port_1,
                                             relative_testcase_path)

    if needs_http:
      # TODO(unassigned): Support https.
      testcase_file_url = testcase_http_url
      testcase_path = testcase_http_url

  # Compose app arguments.
  all_app_args = ''

  if user_profile_argument:
    all_app_args += ' %s=%s' % (user_profile_argument, user_profile_directory)
  if extension_argument and EXTENSIONS_PREFIX in testcase_filename:
    all_app_args += ' %s=%s' % (extension_argument, testcase_directory)
  if apps_argument and APPS_PREFIX in testcase_filename:
    all_app_args += ' %s=%s' % (apps_argument, testcase_directory)
  if window_argument:
    all_app_args += ' %s' % window_argument
  if additional_command_line_flags:
    all_app_args += ' %s' % additional_command_line_flags.strip()
  if app_args:
    all_app_args += ' %s' % app_args.strip()
  # Append %TESTCASE% at end if no testcase pattern is found in app arguments.
  if not utils.sub_string_exists_in(
      ['%TESTCASE%', '%TESTCASE_FILE_URL%', '%TESTCASE_HTTP_URL%'],
      all_app_args) and app_args_append_testcase:
    all_app_args += ' %TESTCASE%'
  all_app_args = all_app_args.strip()

  # Build the actual command to run now.
  if debugger:
    command += '%s ' % debugger
  if app_path:
    command += app_path
  if all_app_args:
    command += ' %s' % all_app_args
  command = command.replace('%APP_DIR%', app_directory)
  command = command.replace('%CRASH_STACKTRACES_DIR%', crash_stacks_directory)
  command = command.replace('%DEVICE_TESTCASES_DIR%',
                            device_testcases_directory)
  command = command.replace('%FUZZER_DIR%', fuzzer_directory)
  command = command.replace('%INPUT_DIR%', input_directory)
  command = command.replace('%ROOT_DIR%', root_directory)
  command = command.replace('%TESTCASE%', testcase_path)
  command = command.replace('%TESTCASE_FILE_URL%', testcase_file_url)
  command = command.replace('%TESTCASE_HTTP_URL%', testcase_http_url)
  command = command.replace('%TMP_DIR%', temp_directory)
  command = command.replace('%USER_PROFILE_DIR%', user_profile_directory)

  if is_android and not launcher:
    # Initial setup phase for command line.
    if write_command_line_file:
      android.adb.write_command_line_file(command, app_path)

    return android.app.get_launch_command(all_app_args, testcase_path,
                                          testcase_file_url)

  # Decide which directory we will run the application from.
  # We are using |app_directory| since it helps to locate pdbs
  # in same directory, other dependencies, etc.
  if os.path.exists(app_directory):
    os.chdir(app_directory)

  return str(command)


def setup_user_profile_directory_if_needed(user_profile_directory):
  """Set user profile directory if it does not exist."""
  if os.path.exists(user_profile_directory):
    # User profile directory already exists. Bail out.
    return

  shell.create_directory(user_profile_directory)

  # Create a file in user profile directory based on format:
  # filename;base64 encoded zlib compressed file contents.
  user_profile_file = environment.get_value('USER_PROFILE_FILE')
  if user_profile_file and ';' in user_profile_file:
    user_profile_filename, encoded_file_contents = (
        user_profile_file.split(';', 1))
    user_profile_file_contents = zlib.decompress(
        base64.b64decode(encoded_file_contents))
    user_profile_file_path = os.path.join(user_profile_directory,
                                          user_profile_filename)
    utils.write_data_to_file(user_profile_file_contents, user_profile_file_path)

  # For Firefox, we need to install a special fuzzPriv extension that exposes
  # special functions to javascript, e.g. gc(), etc.
  app_name = environment.get_value('APP_NAME')
  if app_name.startswith('firefox'):
    # Create extensions directory.
    extensions_directory = os.path.join(user_profile_directory, 'extensions')
    shell.create_directory(extensions_directory)

    # Unpack the fuzzPriv extension.
    extension_archive = os.path.join(environment.get_resources_directory(),
                                     'firefox', 'fuzzPriv-extension.zip')
    archive.unpack(extension_archive, extensions_directory)

    # Add this extension in the extensions configuration file.
    extension_config_file_path = os.path.join(user_profile_directory,
                                              'extensions.ini')
    fuzz_extension_directory = os.path.join(extensions_directory,
                                            'domfuzz@squarefree.com')
    extension_config_file_contents = (
        '[ExtensionDirs]\r\n'
        'Extension0=%s\r\n'
        '\r\n'
        '[ThemeDirs]\r\n' % fuzz_extension_directory)
    utils.write_data_to_file(extension_config_file_contents,
                             extension_config_file_path)


def check_for_bad_build(job_type, crash_revision):
  """Return true if the build is bad, i.e. crashes on startup."""
  # Check the bad build check flag to see if we want do this.
  if not environment.get_value('BAD_BUILD_CHECK'):
    return False

  # Create a blank command line with no file to run and no http.
  command = get_command_line_for_application(file_to_run='', needs_http=False)

  # When checking for bad builds, we use the default window size.
  # We don't want to pick a custom size since it can potentially cause a
  # startup crash and cause a build to be detected incorrectly as bad.
  default_window_argument = environment.get_value('WINDOW_ARG', '')
  if default_window_argument:
    command = command.replace(' %s' % default_window_argument, '')

  # TSAN is slow, and boots slow on first startup. Increase the warmup
  # timeout for this case.
  if environment.tool_matches('TSAN', job_type):
    fast_warmup_timeout = environment.get_value('WARMUP_TIMEOUT')
  else:
    fast_warmup_timeout = environment.get_value('FAST_WARMUP_TIMEOUT')

  # Initialize helper variables.
  is_bad_build = False
  build_run_console_output = ''
  app_directory = environment.get_value('APP_DIR')

  # Exit all running instances.
  process_handler.terminate_stale_application_instances()

  # Check if the build is bad.
  return_code, crash_time, output = process_handler.run_process(
      command,
      timeout=fast_warmup_timeout,
      current_working_directory=app_directory)
  crash_result = CrashResult(return_code, crash_time, output)

  # 1. Need to account for startup crashes with no crash state. E.g. failed to
  #    load shared library. So, ignore state for comparison.
  # 2. Ignore leaks as they don't block a build from reporting regular crashes
  #    and also don't impact regression range calculations.
  if (crash_result.is_crash(ignore_state=True) and
      not crash_result.should_ignore() and
      not crash_result.get_type() in ['Direct-leak', 'Indirect-leak']):
    is_bad_build = True
    build_run_console_output = utils.get_crash_stacktrace_output(
        command,
        crash_result.get_stacktrace(symbolized=True),
        crash_result.get_stacktrace(symbolized=False))
    logs.log(
        'Bad build for %s detected at r%d.' % (job_type, crash_revision),
        output=build_run_console_output)

  # Exit all running instances.
  process_handler.terminate_stale_application_instances()

  # Any of the conditions below indicate that bot is in a bad state and it is
  # not caused by the build itself. In that case, just exit.
  build_state = data_handler.get_build_state(job_type, crash_revision)
  if is_bad_build and utils.sub_string_exists_in(BAD_STATE_HINTS, output):
    logs.log_fatal_and_exit(
        'Bad bot environment detected, exiting.',
        output=build_run_console_output,
        snapshot=process_handler.get_runtime_snapshot())

  # If none of the other bots have added information about this build,
  # then add it now.
  if (build_state == data_types.BuildState.UNMARKED and
      not crash_result.should_ignore()):
    data_handler.add_build_metadata(job_type, crash_revision, is_bad_build,
                                    build_run_console_output)

  return is_bad_build



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers for file operations."""

from clusterfuzz._internal.protos import untrusted_runner_pb2

from . import config


def file_chunk_generator(handle):
  """Yields chunks from handle."""
  data = handle.read(config.FILE_TRANSFER_CHUNK_SIZE)
  while data:
    yield untrusted_runner_pb2.FileChunk(data=data)
    data = handle.read(config.FILE_TRANSFER_CHUNK_SIZE)


def data_chunk_generator(data):
  """Yields chunks for data."""
  index = 0
  while index < len(data):
    cur_chunk = data[index:index + config.FILE_TRANSFER_CHUNK_SIZE]
    yield untrusted_runner_pb2.FileChunk(data=cur_chunk)

    index += config.FILE_TRANSFER_CHUNK_SIZE


def write_chunks(file_path, chunk_iterator):
  """Writes chunks to file."""
  with open(file_path, 'wb') as f:
    for chunk in chunk_iterator:
      f.write(chunk.data)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Cron for checking OSS-Fuzz builds status."""

import datetime
import json
import re

from google.cloud import ndb
import requests

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import handler
from libs import helpers
from libs.issue_management import issue_tracker_utils

BUCKET_URL = 'https://oss-fuzz-build-logs.storage.googleapis.com'
FUZZING_STATUS_URL = BUCKET_URL + '/status.json'
COVERAGE_STATUS_URL = BUCKET_URL + '/status-coverage.json'

FUZZING_BUILD_TYPE = 'fuzzing'
COVERAGE_BUILD_TYPE = 'coverage'
MAIN_BUILD_TYPE = FUZZING_BUILD_TYPE

# It's important not to use a dict, so that fuzzing builds get processed first.
BUILD_STATUS_MAPPINGS = [
    (FUZZING_BUILD_TYPE, FUZZING_STATUS_URL),
    (COVERAGE_BUILD_TYPE, COVERAGE_STATUS_URL),
]

TIMESTAMP_PATTERN = re.compile(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}')
TIMESTAMP_FORMAT = '%Y-%m-%dT%H:%M:%S'

# Consider something wrong if we don't see builds for this many days.
NO_BUILDS_THRESHOLD = datetime.timedelta(days=2)

# Minimum number of consecutive build failures before filing a bug.
MIN_CONSECUTIVE_BUILD_FAILURES = 2

# Number of failures after the last reminder/initial filing to send a reminder.
# This used to be 3 days, but now we may have up to two failures a day since
# https://github.com/google/oss-fuzz/pull/3585 was landed.
REMINDER_INTERVAL = 6


class OssFuzzBuildStatusException(Exception):
  """Exceptions for the build status cron."""


def _get_issue_body(project_name, build_id, build_type):
  """Return the issue body for filing new bugs."""
  template = ('The last {num_builds} builds for {project} have been failing.\n'
              '<b>Build log:</b> {log_link}\n'
              'Build type: {build_type}\n\n'
              'To reproduce locally, please see: '
              'https://google.github.io/oss-fuzz/advanced-topics/reproducing'
              '#reproducing-build-failures\n\n'
              '<b>This bug tracker is not being monitored by OSS-Fuzz team.</b>'
              ' If you have any questions, please create an issue at '
              'https://github.com/google/oss-fuzz/issues/new.\n\n'
              '**This bug will be automatically closed within a '
              'day once it is fixed.**')

  return template.format(
      num_builds=MIN_CONSECUTIVE_BUILD_FAILURES,
      project=project_name,
      log_link=_get_build_link(build_id),
      build_type=build_type)


def _get_oss_fuzz_project(project_name):
  """Return the OssFuzzProject entity for the given project."""
  return ndb.Key(data_types.OssFuzzProject, project_name).get()


def _get_build_link(build_id):
  """Return a link to the build log."""
  return BUCKET_URL + '/log-' + build_id + '.txt'


def _get_ndb_key(project_name, build_type):
  """Constructs a Key literal for build failure entities."""
  if build_type == MAIN_BUILD_TYPE:
    return project_name

  # Use build type suffix for the auxiliary build (e.g. coverage).
  return '%s-%s' % (project_name, build_type)


def create_build_failure(project_name, failure, build_type):
  """Create new build failure."""
  return data_types.OssFuzzBuildFailure(
      id=_get_ndb_key(project_name, build_type),
      project_name=project_name,
      last_checked_timestamp=get_build_time(failure),
      build_type=build_type)


def get_build_failure(project_name, build_type):
  """Return the last build failure for the project."""
  key = ndb.Key(data_types.OssFuzzBuildFailure,
                _get_ndb_key(project_name, build_type))
  return key.get()


def close_build_failure(build_failure):
  """Delete the build failure."""
  build_failure.key.delete()


def get_build_time(build):
  """Return a datetime for when the build was done."""
  # Strip the nanosecond precision from the timestamp, since it's not
  # supported by Python.
  stripped_timestamp = TIMESTAMP_PATTERN.match(build['finish_time'])
  if not stripped_timestamp:
    logs.log_error(
        'Invalid timestamp %s for %s.' % (build['finish_time'], build['name']))
    return None

  return datetime.datetime.strptime(
      stripped_timestamp.group(0), TIMESTAMP_FORMAT)


def file_bug(issue_tracker, project_name, build_id, ccs, build_type):
  """File a new bug for a build failure."""
  logs.log('Filing bug for new build failure (project=%s, build_type=%s, '
           'build_id=%s).' % (project_name, build_type, build_id))

  issue = issue_tracker.new_issue()
  issue.title = '{project_name}: {build_type} build failure'.format(
      project_name=project_name, build_type=build_type.capitalize())
  issue.body = _get_issue_body(project_name, build_id, build_type)
  issue.status = 'New'
  issue.labels.add('Type-Build-Failure')
  issue.labels.add('Proj-' + project_name)

  for cc in ccs:
    issue.ccs.add(cc)

  issue.save()
  return str(issue.id)


def close_bug(issue_tracker, issue_id, project_name):
  """Close a build failure bug."""
  logs.log('Closing build failure bug (project=%s, issue_id=%s).' %
           (project_name, issue_id))

  issue = issue_tracker.get_original_issue(issue_id)
  issue.status = 'Verified'
  issue.save(
      new_comment='The latest build has succeeded, closing this issue.',
      notify=True)


def send_reminder(issue_tracker, issue_id, build_id):
  """Send a reminder about the build still failing."""
  issue = issue_tracker.get_original_issue(issue_id)

  comment = ('Friendly reminder that the the build is still failing.\n'
             'Please try to fix this failure to ensure that fuzzing '
             'remains productive.\n'
             'Latest build log: {log_link}\n')
  comment = comment.format(log_link=_get_build_link(build_id))
  issue.save(new_comment=comment, notify=True)


class Handler(base_handler.Handler):
  """Build status checker."""

  def _close_fixed_builds(self, projects, build_type):
    """Close bugs for fixed builds."""
    issue_tracker = issue_tracker_utils.get_issue_tracker()
    if not issue_tracker:
      raise OssFuzzBuildStatusException('Failed to get issue tracker.')

    for project in projects:
      project_name = project['name']
      builds = project['history']
      if not builds:
        continue

      build_failure = get_build_failure(project_name, build_type)
      if not build_failure:
        continue

      build = builds[0]
      if not build['success']:
        continue

      if build_failure.last_checked_timestamp >= get_build_time(build):
        logs.log_error('Latest successful build time for %s in %s config is '
                       'older than or equal to last failure time.' %
                       (project_name, build_type))
        continue

      if build_failure.issue_id is not None:
        close_bug(issue_tracker, build_failure.issue_id, project_name)

      close_build_failure(build_failure)

  def _process_failures(self, projects, build_type):
    """Process failures."""
    issue_tracker = issue_tracker_utils.get_issue_tracker()
    if not issue_tracker:
      raise OssFuzzBuildStatusException('Failed to get issue tracker.')

    for project in projects:
      project_name = project['name']
      builds = project['history']
      if not builds:
        continue

      build = builds[0]
      if build['success']:
        continue

      project_name = project['name']

      # Do not file an issue for non-main build types, if there is a main build
      # failure for the same project, as the root cause might be the same.
      if build_type != MAIN_BUILD_TYPE:
        build_failure = get_build_failure(project_name, MAIN_BUILD_TYPE)
        if build_failure:
          continue

      build_failure = get_build_failure(project_name, build_type)

      build_time = get_build_time(build)
      if build_failure:
        if build_time <= build_failure.last_checked_timestamp:
          # No updates.
          continue
      else:
        build_failure = create_build_failure(project_name, build, build_type)

      build_failure.last_checked_timestamp = build_time
      build_failure.consecutive_failures += 1
      if build_failure.consecutive_failures >= MIN_CONSECUTIVE_BUILD_FAILURES:
        if build_failure.issue_id is None:
          oss_fuzz_project = _get_oss_fuzz_project(project_name)
          if not oss_fuzz_project:
            logs.log(
                'Project %s is disabled, skipping bug filing.' % project_name)
            continue

          build_failure.issue_id = file_bug(issue_tracker, project_name,
                                            build['build_id'],
                                            oss_fuzz_project.ccs, build_type)
        elif (build_failure.consecutive_failures -
              MIN_CONSECUTIVE_BUILD_FAILURES) % REMINDER_INTERVAL == 0:
          send_reminder(issue_tracker, build_failure.issue_id,
                        build['build_id'])

      build_failure.put()

  def _check_last_get_build_time(self, projects, build_type):
    """Check that builds are up to date."""
    for project in projects:
      project_name = project['name']
      builds = project['history']
      if not builds:
        continue

      build = builds[0]
      time_since_last_build = utils.utcnow() - get_build_time(build)
      if time_since_last_build >= NO_BUILDS_THRESHOLD:
        # Something likely went wrong with the build infrastructure, log errors.
        logs.log_error('%s has not been built in %s config for %d days.' %
                       (project_name, build_type, time_since_last_build.days))

  @handler.cron()
  def get(self):
    """Handles a get request."""
    for build_type, status_url in BUILD_STATUS_MAPPINGS:
      try:
        response = requests.get(status_url)
        response.raise_for_status()
        build_status = json.loads(response.text)
      except (requests.exceptions.RequestException, ValueError) as e:
        raise helpers.EarlyExitException(str(e), response.status_code)

      projects = build_status['projects']

      self._check_last_get_build_time(projects, build_type)
      self._close_fixed_builds(projects, build_type)
      self._process_failures(projects, build_type)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Reproduce tool HTTP utility tests."""
import unittest

from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs.reproduce_tool_fakes import \
    FakeConfig
from clusterfuzz._internal.tests.test_libs.reproduce_tool_fakes import \
    FakeResponse
from clusterfuzz._internal.tests.test_libs.reproduce_tool_fakes import FakeHttp
from local.butler.reproduce_tool import http_utils


class RequestTest(unittest.TestCase):
  """Tests for the request function."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.read_data_from_file',
        'clusterfuzz._internal.base.utils.write_data_to_file',
        'httplib2.Http',
        'local.butler.reproduce_tool.prompts.get_string',
        'webbrowser.open',
    ])

    self.config = FakeConfig()

  def test_multiple_auth_failures(self):
    """Ensure that we don't recurse indefinitely if auth fails persistently."""
    http = FakeHttp([(FakeResponse(401), {}), (FakeResponse(401), {})])
    self.mock.Http.return_value = http
    response, _ = http_utils.request('https://url/', configuration=self.config)

    # Ensure that all expected requests were made.
    self.assertEqual(http.replies, [])

    self.assertEqual(response.status, 401)

  def test_unauthenticated_request(self):
    """Ensure that we can make an unauthenticated request."""
    http = FakeHttp([(FakeResponse(200), {})])
    self.mock.Http.return_value = http
    response, _ = http_utils.request('https://url/', body='test body')

    # Ensure that all expected requests were made.
    self.assertEqual(http.replies, [])

    self.assertEqual(http.last_body, '"test body"')
    self.assertEqual(http.last_headers, {})
    self.assertEqual(response.status, 200)

  def test_authentication(self):
    """Ensure that we can authenticate properly if needed."""
    http = FakeHttp([(FakeResponse(401), {}), (FakeResponse(
        200, include_auth_header=True), {})])
    self.mock.Http.return_value = http
    response, _ = http_utils.request('https://url/', configuration=self.config)

    # Ensure that all expected requests were made.
    self.assertEqual(http.replies, [])

    self.mock.write_data_to_file.assert_called_once_with(
        'fake auth token', http_utils.AUTHORIZATION_CACHE_FILE)
    self.assertEqual(response.status, 200)

  def test_authenticated_request(self):
    """Ensure that we reuse credentials if we've previously authenticated."""
    http = FakeHttp([(FakeResponse(200), {})])
    self.mock.Http.return_value = http
    self.mock.read_data_from_file.return_value = 'cached auth token'
    response, _ = http_utils.request('https://url/', configuration=self.config)

    # Ensure that all expected requests were made.
    self.assertEqual(http.replies, [])

    self.assertEqual(http.last_headers, {
        'Authorization': 'cached auth token',
        'User-Agent': 'clusterfuzz-reproduce'
    })
    self.assertEqual(response.status, 200)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""System log manager related functions."""

import re
import time

from clusterfuzz._internal.metrics import logs

from . import adb


def clear_log():
  """Clear log."""
  adb.run_as_root()
  adb.run_shell_command(['stop', 'logd'])
  adb.run_shell_command(['start', 'logd'])
  time.sleep(0.1)
  adb.run_command(['logcat', '-c'])


def is_line_valid(line):
  """Returns true if we consider this line in logs."""
  if re.match(r'^[-]+ beginning of', line):
    return False

  is_chromium_resource_load = 'NotifyBeforeURLRequest' in line

  # Discard noisy debug and verbose output.
  # http://developer.android.com/tools/debugging/debugging-log.html.
  at_least_info_level = not (line.startswith('D/') or line.startswith('V/'))

  return is_chromium_resource_load or at_least_info_level


def filter_log_output(output):
  """Filters log output. Removes debug info, etc and normalize output."""
  if not output:
    return ''

  filtered_output = ''
  last_process_tuple = (None, None)
  for line in output.splitlines():
    if not is_line_valid(line):
      continue

    # To parse frames like:
    # E/v8      (18890): Error installing extension 'v8/LoadTimes'.
    # {log_level}/{process_name}({process_id}): {message}
    m_line = re.match(r'^[VDIWEFS]/(.+)\(\s*(\d+)\)[:](.*)$', line)
    if not m_line:
      logs.log_error('Failed to parse logcat line: %s' % line)
      continue

    process_name = m_line.group(1).strip()
    process_id = int(m_line.group(2))
    filtered_line = m_line.group(3).rstrip()[1:]

    # Process Android crash stack frames and convert into sanitizer format.
    m_crash_state = re.match(r'\s*#([0-9]+)\s+pc\s+([xX0-9a-fA-F]+)\s+(.+)',
                             filtered_line)
    if m_crash_state:
      frame_no = int(m_crash_state.group(1))
      frame_address = m_crash_state.group(2)
      frame_binary = m_crash_state.group(3).strip()

      # Ignore invalid frames, helps to prevent errors
      # while symbolizing.
      if '<unknown>' in frame_binary:
        continue

      # Normalize frame address.
      if not frame_address.startswith('0x'):
        frame_address = '0x%s' % frame_address

      # Separate out the function argument.
      frame_binary = (frame_binary.split(' '))[0]

      # Normalize line into the same sanitizer tool format.
      filtered_line = ('    #%d %s (%s+%s)' % (frame_no, frame_address,
                                               frame_binary, frame_address))

    # Process Chrome crash stack frames and convert into sanitizer format.
    # Stack frames don't have paranthesis around frame binary and address, so
    # add it explicitly to allow symbolizer to catch it.
    m_crash_state = re.match(
        r'\s*#([0-9]+)\s+([xX0-9a-fA-F]+)\s+([^(]+\+[xX0-9a-fA-F]+)$',
        filtered_line)
    if m_crash_state:
      frame_no = int(m_crash_state.group(1))
      frame_address = m_crash_state.group(2)
      frame_binary_and_address = m_crash_state.group(3).strip()
      filtered_line = ('    #%d %s (%s)' % (frame_no, frame_address,
                                            frame_binary_and_address))

    # Add process number if changed.
    current_process_tuple = (process_name, process_id)
    if current_process_tuple != last_process_tuple:
      filtered_output += '--------- %s (%d):\n' % (process_name, process_id)
      last_process_tuple = current_process_tuple

    filtered_output += filtered_line + '\n'

  return filtered_output


def log_output(additional_flags=''):
  """Return log data without noise and some normalization."""
  output = adb.run_command('logcat -d -v brief %s *:V' % additional_flags)
  return filter_log_output(output)


def log_output_before_last_reboot():
  """Return log data from last reboot without noise and some normalization."""
  return log_output(additional_flags='-L')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for removing duplicate_of of a testcase."""

from flask import request

from handlers import base_handler
from handlers.testcase_detail import show
from libs import handler
from libs import helpers


def remove(testcase):
  """Remove duplicate status from a test case."""
  testcase.status = 'Processed'
  testcase.duplicate_of = None
  testcase.put()

  helpers.log('Removed duplicate status for testcase %s' % testcase.key.id(),
              helpers.MODIFY_OPERATION)


class Handler(base_handler.Handler):
  """Handler that removes duplicate_of of a testcase."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.require_csrf_token
  @handler.check_admin_access
  def post(self):
    """Remove duplicate status from a test case."""
    testcase_id = request.get('testcaseId')
    testcase = helpers.get_testcase(testcase_id)
    remove(testcase)
    return self.render_json(show.get_testcase_detail(testcase))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base classes for other minimizers."""

import copy
import functools
import os
import tempfile
import threading
import time

from clusterfuzz._internal.metrics import logs

from . import errors

DEFAULT_CLEANUP_INTERVAL = 20
DEFAULT_THREAD_COUNT = 8
DEFAULT_TESTS_PER_THREAD = 4

MAX_MERGE_BATCH_SIZE = 32
PROGRESS_REPORT_INTERVAL = 300


class DummyLock(object):
  """Dummy to replace threading.Lock for single-threaded tests."""

  def __enter__(self):
    pass

  def __exit__(self, exec_type, value, traceback):
    pass

  def __bool__(self):
    return False


class TestQueue(object):
  """Queue to store commands that should be executed to test hypotheses."""

  def __init__(self,
               thread_count,
               deadline_check=None,
               progress_report_function=None,
               per_thread_cleanup_function=None):
    self.thread_count = thread_count
    self.deadline_check = deadline_check
    self.progress_report_function = progress_report_function
    self.per_thread_cleanup_function = per_thread_cleanup_function

    self.lock = threading.Lock()
    self.queue = []

  def _pop(self):
    """Pull a single hypothesis to process from the queue."""
    with self.lock:
      if not self.queue:
        return None
      return self.queue.pop(0)

  def _work(self):
    """Process items from the queue until it is empty."""
    while not self.deadline_check or not self.deadline_check(soft_check=True):
      current_item = self._pop()
      if not current_item:
        break

      test, test_function, completion_callback, should_run = current_item  # pylint: disable=unpacking-non-sequence
      if not should_run():
        continue

      result = test_function(test)
      completion_callback(result)

      if self.per_thread_cleanup_function:
        self.per_thread_cleanup_function()

      # Abort if we have exceeded the deadline for this operation.
      if self.deadline_check and self.deadline_check(soft_check=True):
        break

  def _cleanup(self):
    """Clean up the queue to be sure that no more tasks will be executed."""
    with self.lock:
      self.queue = []

  def push(self,
           test,
           test_function,
           completion_callback,
           should_run=lambda: True):
    """Add a test to the queue and a callback to run on completion."""
    with self.lock:
      self.queue.append((test, test_function, completion_callback, should_run))

  def force(self,
            test,
            test_function,
            completion_callback,
            should_run=lambda: True):
    """Force a test to the front of the queue."""
    entry = (test, test_function, completion_callback, should_run)
    with self.lock:
      self.queue.insert(0, entry)

  def size(self):
    """Return the number of unprocessed tasks in the queue."""
    return len(self.queue)

  def process(self):
    """Process all tests in the queue and block until completion."""
    while self.queue:
      threads = [
          threading.Thread(target=self._work) for _ in range(self.thread_count)
      ]
      for thread in threads:
        thread.start()

      while any([thread.is_alive() for thread in threads]):
        if self.deadline_check:
          self.deadline_check(cleanup_function=self._cleanup)

        if self.progress_report_function:
          self.progress_report_function()

        time.sleep(1)


class Testcase(object):
  """Single test case to be minimized."""

  def __init__(self, data, minimizer):
    self.minimizer = minimizer
    if minimizer.tokenize:
      try:
        self.tokens = minimizer.tokenizer(data)
      except UnicodeDecodeError:
        raise errors.AntlrDecodeError
    else:
      self.tokens = data

    self.required_tokens = [True] * len(self.tokens)
    self.tested_hypotheses = set()
    self.unmerged_failing_hypotheses = []
    self.tests_to_queue = []
    self.currently_processing = False
    self.last_progress_report_time = 0
    self.runs_since_last_cleanup = 0
    self.runs_executed = 0

    if minimizer.max_threads > 1:
      self.test_queue = TestQueue(
          minimizer.max_threads,
          deadline_check=self._deadline_exceeded,
          progress_report_function=self._report_progress)
      self.merge_preparation_lock = threading.Lock()
      self.merge_lock = threading.Lock()
      self.cache_lock = threading.Lock()
      self.tests_to_queue_lock = threading.Lock()
    else:
      self.test_queue = None
      self.merge_preparation_lock = DummyLock()
      self.merge_lock = DummyLock()
      self.cache_lock = DummyLock()
      self.tests_to_queue_lock = DummyLock()

  def get_current_testcase_data(self):
    """Return the current test case data."""
    return self.minimizer.token_combiner(self.get_required_tokens())

  # Helper functions based on minimizer configuration.
  def _deadline_exceeded(self, cleanup_function=None, soft_check=False):
    """Check to see if we have exceeded the deadline for execution."""
    if self.minimizer.deadline and time.time() > self.minimizer.deadline:
      if soft_check:
        return True

      # If we are here, we have exceeded the deadline on a hard check. Clean up.
      if cleanup_function:
        cleanup_function()

      if self.minimizer.cleanup_function:
        self.minimizer.cleanup_function()

      # Raise an exception if this is not a soft deadline check.
      raise errors.MinimizationDeadlineExceededError(self)

    return False

  def _delete_file_if_needed(self, input_file):
    """Deletes a temporary file if necessary."""
    # If we are not running in a mode where we need to delete files, do nothing.
    if not self.minimizer.tokenize or not self.minimizer.delete_temp_files:
      return

    try:
      os.remove(input_file)
    except OSError:
      pass

  def _report_progress(self, is_final_progress_report=False):
    """Call a function to report progress if the minimizer uses one."""
    if not self.minimizer.progress_report_function:
      return

    if (time.time() - self.last_progress_report_time < PROGRESS_REPORT_INTERVAL
        and not is_final_progress_report):
      return

    self.last_progress_report_time = time.time()
    message = '%d/%d tokens remaining. %d runs executed so far.' % (len(
        self.get_required_tokens()), len(
            self.required_tokens), self.runs_executed)
    if is_final_progress_report:
      message = "Done with this round of minimization. " + message
    self.minimizer.progress_report_function(message)

  # Functions used when preparing tests.
  def _range_complement(self, current_range):
    """Return required tokens in the complement of the specified range."""
    result = list(range(len(self.tokens)))
    to_remove = set(current_range)
    return [i for i in result if i not in to_remove and self.required_tokens[i]]

  def _prepare_test_input(self, tokens, tested_tokens):
    """Write the tokens currently being tested to a temporary file."""
    tested_tokens = set(tested_tokens)
    current_tokens = [t for i, t in enumerate(tokens) if i in tested_tokens]
    if not self.minimizer.tokenize:
      return current_tokens

    data = self.minimizer.token_combiner(current_tokens)

    handle = self.minimizer.get_temp_file()
    destination = handle.name
    try:
      handle.write(data)
    except IOError:
      # We may have filled the disk. Try processing tests and writing again.
      self._do_single_pass_process()
      handle.write(data)

    handle.close()
    return destination

  def _get_test_file(self, hypothesis):
    """Return a test file for a hypothesis."""
    complement = self._range_complement(hypothesis)
    return self._prepare_test_input(self.tokens, complement)

  def _push_test_to_queue(self, hypothesis):
    """Add a test for a hypothesis to a queue for processing."""
    test_file = self._get_test_file(hypothesis)
    callback = functools.partial(
        self._handle_completed_test,
        hypothesis=hypothesis,
        input_file=test_file)
    should_run = functools.partial(self._contains_required_tokens, hypothesis,
                                   test_file)

    self.test_queue.push(
        test_file,
        self.minimizer.test_function,
        callback,
        should_run=should_run)

    # Make sure that we do not let too many unprocessed tests build up.
    if self.test_queue.size() >= self.minimizer.batch_size:
      self._do_single_pass_process()

  def prepare_test(self, hypothesis):
    """Prepare the test based on the mode we are running in."""
    # Check the cache to make sure we have not tested this before.
    if self._has_tested(hypothesis):
      return

    self.runs_executed += 1
    # If we are single-threaded, just run and process results immediately.
    if not self.test_queue:
      # In the threaded case, we call the cleanup function before each pass
      # over the queue. It needs to be tracked here for the single-thread case.
      self.runs_since_last_cleanup += 1
      if (self.runs_since_last_cleanup >=
          self.minimizer.single_thread_cleanup_interval and
          self.minimizer.cleanup_function):
        self.minimizer.cleanup_function()

      test_file = self._get_test_file(hypothesis)
      if self._contains_required_tokens(hypothesis, test_file):
        self._handle_completed_test(
            self.minimizer.test_function(test_file), hypothesis, test_file)

      # Check to see if we have exceeded the deadline and report progress.
      self._report_progress()
      self._deadline_exceeded()
      return

    if self.currently_processing:
      # If we are processing, we cannot write more tests or add to the queue.
      with self.tests_to_queue_lock:
        self.tests_to_queue.append(hypothesis)
    else:
      self._push_test_to_queue(hypothesis)

  # Functions used when processing test results.
  def _handle_completed_test(self, test_passed, hypothesis, input_file):
    """Update state based on the test result and hypothesis."""
    # If the test failed, handle the result.
    if not test_passed:
      self._handle_failing_hypothesis(hypothesis)

    # Delete leftover files if necessary.
    self._delete_file_if_needed(input_file)

    # Minimizers may need to do something with the test result.
    self._process_test_result(test_passed, hypothesis)

  def _process_test_result(self, test_passed, hypothesis):
    """Additional processing of the result. Minimizers may override this."""

  def _handle_failing_hypothesis(self, hypothesis):
    """Update the token list for a failing hypothesis."""
    if not self.test_queue:
      # We aren't multithreaded, so just update the list directly.
      for token in hypothesis:
        self.required_tokens[token] = False
      return

    with self.merge_preparation_lock:
      self.unmerged_failing_hypotheses.append(hypothesis)
      if len(self.unmerged_failing_hypotheses) < MAX_MERGE_BATCH_SIZE:
        return

      hypotheses_to_merge = self.unmerged_failing_hypotheses
      self.unmerged_failing_hypotheses = []

    # We may need to block while the previous batch is merging. If not, the
    # results from this batch could conflict with the results from the previous.
    with self.merge_lock:
      self._attempt_merge(hypotheses_to_merge)

  def _attempt_merge(self, hypotheses):
    """Update the required token list if the queued changes don't conflict."""
    # If there's nothing to merge, we're done.
    if not hypotheses:
      return

    aggregate_tokens = set()
    for hypothesis in hypotheses:
      for token in hypothesis:
        aggregate_tokens.add(token)
    aggregate_hypothesis = list(aggregate_tokens)

    complement = self._range_complement(aggregate_hypothesis)
    test_file = self._prepare_test_input(self.tokens, complement)
    test_passed = self.minimizer.test_function(test_file)
    self._delete_file_if_needed(test_file)

    # Failed (crashed), so there was no conflict here.
    if not test_passed:
      for token in aggregate_hypothesis:
        self.required_tokens[token] = False
      return

    # Passed (no crash). We need to try a bit harder to resolve this conflict.
    if len(hypotheses) == 1:
      # We really cannot remove this token. No additional work to be done.
      return

    middle = len(hypotheses) // 2
    front = hypotheses[:middle]
    back = hypotheses[middle:]

    # This could potentially be optimized to assume that if one test fails the
    # other would pass, but because of flaky tests it's safer to run the test
    # unconditionally.
    self._attempt_merge(front)
    self._attempt_merge(back)

  def _do_single_pass_process(self):
    """Process through a single pass of our test queue."""
    self.currently_processing = True
    self.test_queue.process()

    # If a cleanup function is provided, call it. This is usually used to
    # ensure that all processes are terminated or perform additional cleanup.
    if self.minimizer.cleanup_function:
      self.minimizer.cleanup_function()

    # Push any results generated while this test was running to the queue.
    self.currently_processing = False
    while self.tests_to_queue:
      with self.tests_to_queue_lock:
        hypothesis = self.tests_to_queue.pop(0)

      # This may trigger another round of processing, so don't hold the lock.
      self._push_test_to_queue(hypothesis)

  def process(self):
    """Start a test."""
    if not self.test_queue:
      return

    while self.test_queue.size():
      self._do_single_pass_process()

    with self.merge_preparation_lock:
      hypotheses_to_merge = self.unmerged_failing_hypotheses
      self.unmerged_failing_hypotheses = []

    with self.merge_lock:
      self._attempt_merge(hypotheses_to_merge)

  # Cache functions.
  def _contains_required_tokens(self, hypothesis, test_file):
    """Check to see if this hypothesis contains untested tokens."""
    # It is possible that we could copy this while it is being updated. We do
    # not block in this case because the worst case scenario is that we run an
    # irrelevant test, and blocking is potentially expensive.
    working_required_tokens = copy.copy(self.required_tokens)
    with self.merge_preparation_lock:
      # A deep copy is not required. Hypotheses are not modified after being
      # added to the list for processing.
      unprocessed_hypotheses = copy.copy(self.unmerged_failing_hypotheses)

    for unprocessed_hypothesis in unprocessed_hypotheses:
      for token in unprocessed_hypothesis:
        # For this check, we do not care if the merge would succeed or not since
        # the best case is that we would add the token to the queue as well.
        working_required_tokens[token] = False

    for token in hypothesis:
      if working_required_tokens[token]:
        return True

    # If we aren't going to run this test, this will not have a completion
    # callback. If that happens, we need to clean up now.
    self._delete_file_if_needed(test_file)
    return False

  def _has_tested(self, hypothesis):
    """Check to see if this hypothesis has been tested before."""
    hypothesis_tuple = tuple(hypothesis)
    with self.cache_lock:
      if hypothesis_tuple in self.tested_hypotheses:
        return True

      self.tested_hypotheses.add(hypothesis_tuple)

    return False

  # Result checking functions.
  def get_result(self):
    """Get the result of minimization."""
    # Done with minimization, output log one more time
    self._report_progress(is_final_progress_report=True)

    if not self.minimizer.tokenize:
      return self.get_required_tokens()
    return self.get_current_testcase_data()

  def get_required_tokens(self):
    """Return all required tokens for this test case."""
    return [t for i, t in enumerate(self.tokens) if self.required_tokens[i]]

  def get_required_token_indices(self):
    """Get the indices of all remaining required tokens."""
    return [i for i, v in enumerate(self.required_tokens) if v]


def _default_tokenizer(s):
  """Default string tokenizer which splits on newlines."""
  return s.split(b'\n')


def _default_combiner(tokens):
  """Default token combiner which assumes each token is a line."""
  return b'\n'.join(tokens)


class Minimizer(object):
  """Base class for minimizers."""

  def __init__(self,
               test_function,
               max_threads=1,
               tokenizer=_default_tokenizer,
               token_combiner=_default_combiner,
               tokenize=True,
               cleanup_function=None,
               single_thread_cleanup_interval=DEFAULT_CLEANUP_INTERVAL,
               deadline=None,
               get_temp_file=None,
               delete_temp_files=True,
               batch_size=None,
               progress_report_function=None,
               file_extension=''):
    """Initialize a minimizer. A minimizer object can be used multiple times."""
    self.test_function = test_function
    self.max_threads = max_threads
    self.tokenizer = tokenizer
    self.token_combiner = token_combiner
    self.tokenize = tokenize
    self.cleanup_function = cleanup_function
    self.single_thread_cleanup_interval = single_thread_cleanup_interval
    self.deadline = deadline
    self.get_temp_file = get_temp_file
    self.delete_temp_files = delete_temp_files
    self.progress_report_function = progress_report_function

    if batch_size:
      self.batch_size = batch_size
    else:
      self.batch_size = DEFAULT_TESTS_PER_THREAD * max_threads

    if not get_temp_file:
      self.get_temp_file = functools.partial(
          tempfile.NamedTemporaryFile,
          mode='wb',
          delete=False,
          prefix='min_',
          suffix=file_extension)
    else:
      self.get_temp_file = get_temp_file

  @staticmethod
  def _handle_constructor_argument(key, kwargs, default=None):
    """Cleanup a keyword argument specific to a subclass and get the value."""
    result = default
    try:
      result = kwargs[key]
      del kwargs[key]
    except KeyError:
      pass

    return result

  def _execute(self, data):
    """Perform minimization on a test case."""
    raise NotImplementedError

  def minimize(self, data):
    """Wrapper to perform common tasks and call |_execute|."""
    try:
      testcase = self._execute(data)
    except errors.MinimizationDeadlineExceededError as error:
      # When a MinimizationDeadlineExceededError is raised, the partially
      # minimized test case is stored with it so that we can recover the work
      # that had been done up to that point.
      testcase = error.testcase
    except errors.TokenizationFailureError:
      logs.log('Tokenized data did not match original data. Defaulting to line'
               'minimization.')
      # In situation where the tokenizer does not work, we still want to use
      # the token combiner. This will not change the data unless
      # token combiner changes the data such as appending extra data to the
      # start or end. If this is the case, that change will be expected
      # in the return.
      return self.token_combiner([data])

    return testcase.get_result()

  def validate_tokenizer(self, data, testcase):
    """Validate that the tokenizer correctly tokenized the data. This is
    necessary because if the tokenizer does not recognize a character, it will
    skip it."""
    # If data is a list, it means we're not minimizing a test case but another
    # feature such as files or command line arguments. In these cases, we don't
    # rely on a tokenizer.
    if isinstance(data, list):
      return True

    # For most token_combiners, using the combiner on data like below will do
    # nothing, but in situations where data is changed in the token combiner
    # such as data being appended to the start or end of data we want to make
    # sure the same change happens to both before comparison.
    data = self.token_combiner([data])
    return testcase.get_current_testcase_data() == data

  @staticmethod
  def run(data, thread_count=DEFAULT_THREAD_COUNT, file_extension=''):
    """Minimize |data| using this minimizer's default configuration."""
    raise NotImplementedError



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for marking a testcase as security-related."""

from flask import request

from clusterfuzz._internal.base import bisection
from clusterfuzz._internal.crash_analysis import severity_analyzer
from handlers import base_handler
from handlers.testcase_detail import show
from libs import access
from libs import handler
from libs import helpers


def mark(testcase, security, severity):
  """Mark the testcase as security-related."""
  testcase.security_flag = security
  if security:
    if not severity:
      severity = severity_analyzer.get_security_severity(
          testcase.crash_type, testcase.crash_stacktrace, testcase.job_type,
          bool(testcase.gestures))

    testcase.security_severity = severity
    bisection.request_bisection(testcase)
  else:
    # The bisection infrastructure only cares about security bugs. If this was
    # marked as non-security, mark it as invalid.
    bisection.notify_bisection_invalid(testcase)

  testcase.put()
  helpers.log(
      f'Set security flags on testcase {testcase.key.id()} to {security}.',
      helpers.MODIFY_OPERATION)


class Handler(base_handler.Handler):
  """Handler that removes an issue from a testcase."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.require_csrf_token
  def post(self):
    """Mark the testcase as security-related."""
    testcase_id = request.get('testcaseId')
    security = request.get('security')
    severity = request.get('severity')
    testcase = helpers.get_testcase(testcase_id)

    if not access.has_access(
        fuzzer_name=testcase.actual_fuzzer_name(),
        job_type=testcase.job_type,
        need_privileged_access=True):
      raise helpers.AccessDeniedException()

    mark(testcase, security, severity)
    return self.render_json(show.get_testcase_detail(testcase))
